# stdlib

# third party
from sympc.tensor import ShareTensor

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.torch.tensor_util import protobuf_tensor_deserializer
from ...lib.torch.tensor_util import protobuf_tensor_serializer
from ...proto.lib.sympc.share_tensor_pb2 import ShareTensor as ShareTensor_PB
from .session_util import protobuf_session_deserializer
from .session_util import protobuf_session_serializer


def object2proto(obj: object) -> ShareTensor_PB:

    share: ShareTensor = obj

    session = protobuf_session_serializer(share.session)
    proto = ShareTensor_PB(session=session)

    tensor_data = getattr(share.tensor, "data", None)
    if tensor_data is not None:
        proto.tensor.tensor.CopyFrom(protobuf_tensor_serializer(tensor_data))
    proto.tensor.requires_grad = getattr(share.tensor, "requires_grad", False)
    grad = getattr(share.tensor, "grad", None)
    if grad is not None:
        proto.tensor.grad.CopyFrom(protobuf_tensor_serializer(grad))

    return proto


def proto2object(proto: ShareTensor_PB) -> ShareTensor:
    session = protobuf_session_deserializer(proto=proto.session)

    data = protobuf_tensor_deserializer(proto.tensor.tensor)
    share = ShareTensor(data=None, session=session)

    # Manually put the tensor since we do not want to re-encode it
    share.tensor = data.type(session.tensor_type)

    return share


GenerateWrapper(
    wrapped_type=ShareTensor,
    import_path="sympc.tensor.ShareTensor",
    protobuf_scheme=ShareTensor_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# third party
import pytest

# syft absolute
import syft as sy
from syft import serialize
from syft.core.node.common.service.child_node_lifecycle_service import (
    RegisterChildNodeMessage,
)


@pytest.mark.slow
def test_child_node_lifecycle_message_serde(
    node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    second_client = node.get_client()

    # bob_phone_client.register(client=bob_vm_client)
    # generates this message
    msg = RegisterChildNodeMessage(
        lookup_id=client.id,  # TODO: not sure if this is needed anymore
        child_node_client_address=client.address,
        address=second_client.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == msg2.address
    assert msg.child_node_client_address == msg2.child_node_client_address
    assert msg == msg2



# stdlib
from abc import ABC
from collections import OrderedDict
from collections import UserDict
from collections import UserList
from collections import UserString
from typing import Any
from typing import Optional
from typing import Union

# syft relative
from .. import python
from ...core.common import UID
from ...logger import traceback_and_raise
from .primitive_interface import PyPrimitive

NoneType = type(None)

primitives = [
    bool,
    dict,
    complex,
    float,
    int,
    list,
    tuple,
    set,
    slice,
    None,
    NoneType,
    str,
    UserDict,
    UserList,
    UserString,
    OrderedDict,
]

PrimitiveType = Union[
    bool,
    dict,
    complex,
    float,
    int,
    tuple,
    list,
    set,
    slice,
    None,
    NoneType,
    str,
    UserDict,
    UserList,
    UserString,
    OrderedDict,
]


def isprimitive(value: Any) -> bool:
    if not issubclass(type(value), PyPrimitive) and type(value) in primitives:
        return True
    return False


class PrimitiveFactory(ABC):
    def upcast(self) -> Union[int, float, bool, complex, list, str, None]:
        traceback_and_raise(NotImplementedError)

    @staticmethod
    def generate_primitive(
        value: Union[PrimitiveType, type(NotImplemented), PyPrimitive],  # type: ignore
        id: Optional[UID] = None,
        recurse: bool = False,
    ) -> Any:
        if isinstance(value, PyPrimitive):
            return value

        if isinstance(value, bool):
            return python.Bool(value=value, id=id)

        if isinstance(value, int):
            return python.Int(value=value, id=id)

        if isinstance(value, float):
            return python.Float(value=value, id=id)

        if isinstance(value, complex):
            return python.Complex(real=value.real, imag=value.imag, id=id)

        if isinstance(value, tuple):
            return python.Tuple(value)

        if isinstance(value, set):
            return python.Set(value)

        if isinstance(value, slice):
            return python.Slice(
                start=value.start, stop=value.stop, step=value.step, id=id
            )

        if type(value) in [list, UserList]:
            if not recurse:
                return python.List(value=value, id=id)
            else:
                # allow recursive primitive downcasting
                new_list = []
                if value is not None:
                    for val in value:
                        if isprimitive(value=val):
                            new_list.append(
                                PrimitiveFactory.generate_primitive(
                                    value=val, recurse=recurse
                                )
                            )
                        else:
                            new_list.append(val)
                return python.List(value=new_list, id=id)

        if type(value) in [dict, UserDict, OrderedDict]:
            constructor = (
                python.collections.OrderedDict
                if type(value) is OrderedDict
                else python.Dict
            )

            if not recurse:
                new_dict = constructor(value)
            else:
                # allow recursive primitive downcasting
                new_dict = constructor()
                if value is not None:
                    items = getattr(value, "items", None)
                    if items is not None:
                        for k, val in items():
                            if isprimitive(value=val):
                                new_dict[k] = PrimitiveFactory.generate_primitive(
                                    value=val, recurse=recurse
                                )
                            else:
                                new_dict[k] = val
            # if we pass id in as a kwargs it ends up in the actual dict
            if id is not None:
                new_dict._id = id
            return new_dict

        if type(value) in [str, UserString]:
            return python.String(value=value, id=id)

        if value is NotImplemented:
            return value

        return python.SyNone



# stdlib
from functools import partial

# third party
import pytest
from pytest import CaptureFixture

# syft absolute
import syft
from syft.ast.globals import Globals
from syft.core.node.common.client import Client
from syft.lib import lib_ast

# syft relative
from . import module_test
from .functionality_test import module_test_methods
from .functionality_test import update_ast_test

iter_without_len_methods = [
    ("module_test.IterWithoutLen", "module_test.IterWithoutLen"),
    ("module_test.IterWithoutLen.__iter__", "syft.lib.python.Iterator"),
    ("module_test.IterWithoutLen.__next__", "syft.lib.python.Int"),
]


@pytest.fixture(scope="function")
def register_module_test_iter_without_len() -> None:
    # Make lib_ast contain the specific methods/attributes
    update_ast_test(ast_or_client=syft.lib_ast, methods=iter_without_len_methods)

    # Make sure that when we register a new client it would update the specific AST
    lib_ast.loaded_lib_constructors["module_test"] = partial(
        update_ast_test, methods=iter_without_len_methods
    )


@pytest.fixture()
def custom_client() -> Client:
    alice = syft.VirtualMachine(name="alice")
    alice_client = alice.get_root_client()

    return alice_client


# -------------------- Module Tests --------------------


def test_module_repr() -> None:
    ast = Globals(None)

    for method, return_type in module_test_methods:
        ast.add_path(
            path=method, framework_reference=module_test, return_type_name=return_type
        )

    expected_repr = "Module:\n"
    for name, module in ast.attrs.items():
        expected_repr += (
            "\t." + name + " -> " + str(module).replace("\t.", "\t\t.") + "\n"
        )

    assert ast.__repr__() == expected_repr


# -------------------- Klass Tests --------------------


def test_klass_get_and_set_request_config() -> None:
    ast = Globals(None)

    for method, return_type in module_test_methods:
        ast.add_path(
            path=method, framework_reference=module_test, return_type_name=return_type
        )

    for klass in ast.classes:
        klass.create_pointer_class()
        klass_ptr = klass.pointer_type

        # Test Klass pointer has get_request_config attribute
        get_request_config = klass.pointer_type.get_request_config
        assert get_request_config(klass_ptr) == {
            "request_block": True,
            "timeout_secs": 25,
            "delete_obj": False,
        }

        # Test Klass pointer has set_request_config attribute and it works
        set_request_config = klass.pointer_type.set_request_config
        set_request_config(
            klass_ptr,
            {
                "request_block": False,
                "timeout_secs": 25,
                "delete_obj": True,
            },
        )
        get_request_config = klass.pointer_type.get_request_config
        assert get_request_config() == {
            "request_block": False,
            "timeout_secs": 25,
            "delete_obj": True,
        }


def test_klass_wrap_iterator_raises_exception(
    register_module_test_iter_without_len: CaptureFixture, custom_client: Client
) -> None:
    iter_without_len_ptr = custom_client.module_test.IterWithoutLen()
    with pytest.raises(ValueError) as exception_info:
        iter_without_len_ptr.__iter__()
    assert (
        str(exception_info.value)
        == "Can't build a remote iterator on an object with no __len__."
    )



# stdlib
import secrets
from typing import Tuple

# syft absolute
import syft as sy
from syft import serialize
from syft.core.common.message import SyftMessage
from syft.core.io.address import Address
from syft.core.node.common.node import Node
from syft.grid.services.signaling_service import AnswerPullRequestMessage
from syft.grid.services.signaling_service import InvalidLoopBackRequest
from syft.grid.services.signaling_service import OfferPullRequestMessage
from syft.grid.services.signaling_service import PullSignalingService
from syft.grid.services.signaling_service import PushSignalingService
from syft.grid.services.signaling_service import RegisterDuetPeerService
from syft.grid.services.signaling_service import RegisterNewPeerMessage
from syft.grid.services.signaling_service import SignalingAnswerMessage
from syft.grid.services.signaling_service import SignalingOfferMessage
from syft.grid.services.signaling_service import SignalingRequestsNotFound


def get_preset_nodes() -> Tuple[Node, Node, Node]:
    om_network = sy.Network(name="OpenMined")
    om_network.immediate_services_without_reply.append(PushSignalingService)
    om_network.immediate_services_with_reply.append(PullSignalingService)
    om_network.immediate_services_with_reply.append(RegisterDuetPeerService)
    om_network._register_services()  # re-register all services including SignalingService
    bob_vm = sy.VirtualMachine(name="Bob")
    alice_vm = sy.VirtualMachine(name="Alice")
    return om_network, bob_vm, alice_vm


def test_signaling_offer_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    target_id = secrets.token_hex(nbytes=16)
    host_id = secrets.token_hex(nbytes=16)

    msg = SignalingOfferMessage(
        address=target,
        payload="SDP",
        host_metadata=node.get_metadata_for_client(),
        target_peer=target_id,
        host_peer=host_id,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    msg_metadata = node.get_metadata_for_client()
    msg2_metadata = msg2.host_metadata

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.payload == msg2.payload
    assert msg2.payload == "SDP"
    assert msg2.host_peer == host_id
    assert msg2.target_peer == target_id
    assert msg == msg2

    assert msg_metadata.name == msg2_metadata.name
    assert msg_metadata.node == msg2_metadata.node
    assert msg_metadata.id == msg2_metadata.id


def test_signaling_answer_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")
    target_id = secrets.token_hex(nbytes=16)
    host_id = secrets.token_hex(nbytes=16)

    msg = SignalingAnswerMessage(
        address=target,
        payload="SDP",
        host_metadata=node.get_metadata_for_client(),
        target_peer=target_id,
        host_peer=host_id,
    )
    msg_metadata = node.get_metadata_for_client()

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)
    msg2_metadata = msg2.host_metadata

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.payload == msg2.payload
    assert msg2.payload == "SDP"
    assert msg2.host_peer == host_id
    assert msg2.target_peer == target_id
    assert msg == msg2

    assert msg_metadata.name == msg2_metadata.name
    assert msg_metadata.node == msg2_metadata.node
    assert msg_metadata.id == msg2_metadata.id


def test_signaling_answer_pull_request_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    target_id = secrets.token_hex(nbytes=16)
    host_id = secrets.token_hex(nbytes=16)

    msg = AnswerPullRequestMessage(
        address=target,
        target_peer=target_id,
        host_peer=host_id,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg == msg2
    assert msg2.host_peer == host_id
    assert msg2.target_peer == target_id


def test_signaling_offer_pull_request_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    target_id = secrets.token_hex(nbytes=16)
    host_id = secrets.token_hex(nbytes=16)

    msg = OfferPullRequestMessage(
        address=target,
        target_peer=target_id,
        host_peer=host_id,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg == msg2
    assert msg2.host_peer == host_id
    assert msg2.target_peer == target_id


def test_push_offer_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    target_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id
    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    offer_msg = SignalingOfferMessage(
        address=om_network.address,
        payload="SDP",
        host_metadata=alice_vm.get_metadata_for_client(),
        target_peer=target_id,
        host_peer=host_id,
    )
    om_network_client.send_immediate_msg_without_reply(msg=offer_msg)

    assert (
        om_network.signaling_msgs[target_id][SyftMessage].pop(offer_msg.id) == offer_msg
    )


def test_push_answer_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    target_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id
    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    offer_msg = SignalingAnswerMessage(
        address=om_network.address,
        payload="SDP",
        host_metadata=alice_vm.get_metadata_for_client(),
        target_peer=target_id,
        host_peer=host_id,
    )
    om_network_client.send_immediate_msg_without_reply(msg=offer_msg)

    assert (
        om_network.signaling_msgs[target_id][SyftMessage].pop(offer_msg.id) == offer_msg
    )


def test_pull_offer_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    target_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id
    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    offer_msg = SignalingOfferMessage(
        address=om_network.address,
        payload="SDP",
        host_metadata=alice_vm.get_metadata_for_client(),
        target_peer=target_id,
        host_peer=host_id,
    )
    om_network_client.send_immediate_msg_without_reply(msg=offer_msg)

    assert (
        om_network.signaling_msgs[target_id][SyftMessage].get(offer_msg.id) == offer_msg
    )

    offer_pull_req = OfferPullRequestMessage(
        address=om_network.address,
        target_peer=host_id,
        host_peer=target_id,
        reply_to=om_network_client.address,
    )

    response = om_network_client.send_immediate_msg_with_reply(msg=offer_pull_req)

    assert response == offer_msg


def test_pull_answer_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    target_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id
    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    answer_msg = SignalingAnswerMessage(
        address=om_network.address,
        payload="SDP",
        host_metadata=alice_vm.get_metadata_for_client(),
        target_peer=target_id,
        host_peer=host_id,
    )
    om_network_client.send_immediate_msg_without_reply(msg=answer_msg)

    assert (
        om_network.signaling_msgs[target_id][SyftMessage].get(answer_msg.id)
        == answer_msg
    )

    ans_pull_req = AnswerPullRequestMessage(
        address=om_network.address,
        target_peer=host_id,
        host_peer=target_id,
        reply_to=om_network_client.address,
    )

    response = om_network_client.send_immediate_msg_with_reply(msg=ans_pull_req)

    assert response == answer_msg


def test_loopback_offer_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    offer_msg = SignalingOfferMessage(
        address=om_network.address,
        payload="SDP",
        host_metadata=alice_vm.get_metadata_for_client(),
        target_peer=host_id,
        host_peer=host_id,
    )

    om_network_client.send_immediate_msg_without_reply(msg=offer_msg)

    # Do not enqueue loopback requests
    assert len(om_network.signaling_msgs[host_id][SyftMessage]) == 0

    offer_pull_req = OfferPullRequestMessage(
        address=om_network.address,
        target_peer=host_id,
        host_peer=host_id,
        reply_to=om_network_client.address,
    )

    response = om_network_client.send_immediate_msg_with_reply(msg=offer_pull_req)

    # Return InvalidLoopBack Message
    assert isinstance(response, InvalidLoopBackRequest)


def test_loopback_answer_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    metadata = alice_vm.get_metadata_for_client()
    answer_msg = SignalingAnswerMessage(
        address=om_network.address,
        payload="SDP",
        host_metadata=metadata,
        target_peer=host_id,
        host_peer=host_id,
    )
    om_network_client.send_immediate_msg_without_reply(msg=answer_msg)

    # Do not enqueue loopback requests
    assert len(om_network.signaling_msgs[host_id][SyftMessage]) == 0

    ans_pull_req = AnswerPullRequestMessage(
        address=om_network.address,
        target_peer=host_id,
        host_peer=host_id,
        reply_to=om_network_client.address,
    )

    response = om_network_client.send_immediate_msg_with_reply(msg=ans_pull_req)

    # Return InvalidLoopBackMessage
    assert isinstance(response, InvalidLoopBackRequest)


def test_not_found_pull_offer_requests_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id
    target_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    offer_pull_req = OfferPullRequestMessage(
        address=om_network.address,
        target_peer=target_id,
        host_peer=host_id,
        reply_to=om_network_client.address,
    )

    response = om_network_client.send_immediate_msg_with_reply(msg=offer_pull_req)

    assert isinstance(response, SignalingRequestsNotFound)


def test_not_found_pull_ans_requests_signaling_service() -> None:
    om_network, bob_vm, alice_vm = get_preset_nodes()
    om_network_client = om_network.get_root_client()

    msg = RegisterNewPeerMessage(
        address=om_network.address, reply_to=om_network_client.address
    )

    host_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id
    target_id = om_network_client.send_immediate_msg_with_reply(msg=msg).peer_id

    ans_pull_req = AnswerPullRequestMessage(
        address=om_network.address,
        target_peer=target_id,
        host_peer=host_id,
        reply_to=om_network_client.address,
    )

    response = om_network_client.send_immediate_msg_with_reply(msg=ans_pull_req)

    assert isinstance(response, SignalingRequestsNotFound)



# flake8: noqa
"""
Tests copied from cpython test suite:
https://github.com/python/cpython/blob/3.9/Lib/test/test_dict.py
"""

# stdlib
import collections
import collections.abc
import gc
import pickle
import random
import string
import sys
from test import support
import unittest
import weakref

# third party
import pytest

# syft absolute
from syft.lib.python.dict import Dict
from syft.lib.python.none import SyNone
from syft.lib.python.string import String

# import weakref


class DictTest(unittest.TestCase):
    def test_invalid_keyword_arguments(self):
        class Custom(dict):
            pass

        for invalid in {1: 2}, Custom({1: 2}):
            with self.assertRaises(TypeError):
                dict(**invalid)
            with self.assertRaises(TypeError):
                {}.update(**invalid)

    def test_constructor(self):
        # calling built-in types without argument must return empty
        self.assertEqual(Dict(), {})
        self.assertIsNot(Dict(), {})

    @pytest.mark.slow
    def test_literal_constructor(self):
        # check literal constructor for different sized dicts
        # (to exercise the BUILD_MAP oparg).
        for n in (0, 1, 6, 256, 400):
            items = [
                ("".join(random.sample(string.ascii_letters, 8)), i) for i in range(n)
            ]
            random.shuffle(items)
            formatted_items = (f"{k!r}: {v:d}" for k, v in items)
            dictliteral = "{" + ", ".join(formatted_items) + "}"
            self.assertEqual(eval(dictliteral), dict(items))

    def test_merge_operator(self):

        a = Dict({0: 0, 1: 1, 2: 1})
        b = Dict({1: 1, 2: 2, 3: 3})

        if sys.version_info >= (3, 9):
            c = a.copy()
            c |= b

            self.assertEqual(a | b, Dict({0: 0, 1: 1, 2: 2, 3: 3}))
            self.assertEqual(c, Dict({0: 0, 1: 1, 2: 2, 3: 3}))

            c = b.copy()
            c |= a

            self.assertEqual(b | a, Dict({1: 1, 2: 1, 3: 3, 0: 0}))
            self.assertEqual(c, Dict({1: 1, 2: 1, 3: 3, 0: 0}))

            c = a.copy()
            c |= [(1, 1), (2, 2), (3, 3)]

            self.assertEqual(c, Dict({0: 0, 1: 1, 2: 2, 3: 3}))

            self.assertIs(a.__or__(None), NotImplemented)
            self.assertIs(a.__or__(()), NotImplemented)
            self.assertIs(a.__or__("BAD"), NotImplemented)
            self.assertIs(a.__or__(""), NotImplemented)

            self.assertRaises(TypeError, a.__ior__, None)
            self.assertEqual(a.__ior__(()), {0: 0, 1: 1, 2: 1})
            self.assertRaises(ValueError, a.__ior__, "BAD")
            self.assertEqual(a.__ior__(""), {0: 0, 1: 1, 2: 1})

    def test_bool(self):
        self.assertIs(not {}, True)
        self.assertTrue(Dict({1: 2}))
        self.assertIs(bool(Dict({})), False)
        self.assertIs(bool(Dict({1: 2})), True)

    def test_keys(self):
        d = Dict()
        self.assertEqual(set(d.keys()), set())
        d = {"a": 1, "b": 2}
        k = d.keys()
        self.assertEqual(set(k), {"a", "b"})
        self.assertIn("a", k)
        self.assertIn("b", k)
        self.assertIn("a", d)
        self.assertIn("b", d)
        self.assertRaises(TypeError, d.keys, None)
        self.assertEqual(repr(dict(a=1).keys()), "dict_keys(['a'])")

    def test_values(self):
        d = Dict()
        self.assertEqual(set(d.values()), set())
        d = Dict({1: 2})
        self.assertEqual(set(d.values()), {2})
        self.assertRaises(TypeError, d.values, None)
        self.assertEqual(repr(dict(a=1).values()), "dict_values([1])")

    @pytest.mark.xfail
    def test_items(self):
        # TODO: support this when we have sets:
        d = Dict()
        self.assertEqual(set(d.items()), set())

        d = Dict({1: 2})
        self.assertEqual(set(d.items()), {(1, 2)})
        self.assertRaises(TypeError, d.items, None)
        self.assertEqual(repr(dict(a=1).items()), "dict_items([('a', 1)])")

    def test_contains(self):
        d = Dict()
        self.assertNotIn("a", d)
        self.assertFalse("a" in d)
        self.assertTrue("a" not in d)
        d = Dict({"a": 1, "b": 2})
        self.assertIn("a", d)
        self.assertIn("b", d)
        self.assertNotIn("c", d)

        self.assertRaises(TypeError, d.__contains__)

    def test_len(self):
        d = Dict()
        self.assertEqual(len(d), 0)
        d = Dict({"a": 1, "b": 2})
        self.assertEqual(len(d), 2)

    def test_getitem(self):
        d = Dict({"a": 1, "b": 2})
        self.assertEqual(d["a"], 1)
        self.assertEqual(d["b"], 2)
        d["c"] = 3
        d["a"] = 4
        self.assertEqual(d["c"], 3)
        self.assertEqual(d["a"], 4)
        del d["b"]
        self.assertEqual(d, {"a": 4, "c": 3})

        self.assertRaises(TypeError, d.__getitem__)

        class BadEq(object):
            def __eq__(self, other):
                raise Exc()

            def __hash__(self):
                return 24

        d = Dict()
        d[BadEq()] = 42
        self.assertRaises(KeyError, d.__getitem__, 23)

        class Exc(Exception):
            pass

        class BadHash(object):
            fail = False

            def __hash__(self):
                if self.fail:
                    raise Exc()
                else:
                    return 42

        x = BadHash()
        d[x] = 42
        x.fail = True
        self.assertRaises(Exc, d.__getitem__, x)

    def test_clear(self):
        d = Dict({1: 1, 2: 2, 3: 3})
        d.clear()
        self.assertEqual(d, {})

        self.assertRaises(TypeError, d.clear, None)

    def test_update(self):
        d = Dict()
        d.update({1: 100})
        d.update({2: 20})
        d.update({1: 1, 2: 2, 3: 3})
        self.assertEqual(d, {1: 1, 2: 2, 3: 3})

        d.update()
        self.assertEqual(d, {1: 1, 2: 2, 3: 3})

        self.assertRaises((TypeError, AttributeError), d.update, None)

        class SimpleUserDict:
            def __init__(self):
                self.d = {1: 1, 2: 2, 3: 3}

            def keys(self):
                return self.d.keys()

            def __getitem__(self, i):
                return self.d[i]

        d.clear()
        d.update(SimpleUserDict())
        self.assertEqual(d, {1: 1, 2: 2, 3: 3})

        class Exc(Exception):
            pass

        d.clear()

        class FailingUserDict:
            def keys(self):
                raise Exc

        self.assertRaises(Exc, d.update, FailingUserDict())

        class FailingUserDict:
            def keys(self):
                class BogonIter:
                    def __init__(self):
                        self.i = 1

                    def __iter__(self):
                        return self

                    def __next__(self):
                        if self.i:
                            self.i = 0
                            return "a"
                        raise Exc

                return BogonIter()

            def __getitem__(self, key):
                return key

        self.assertRaises(Exc, d.update, FailingUserDict())

        class FailingUserDict:
            def keys(self):
                class BogonIter:
                    def __init__(self):
                        self.i = ord("a")

                    def __iter__(self):
                        return self

                    def __next__(self):
                        if self.i <= ord("z"):
                            rtn = chr(self.i)
                            self.i += 1
                            return rtn
                        raise StopIteration

                return BogonIter()

            def __getitem__(self, key):
                raise Exc

        self.assertRaises(Exc, d.update, FailingUserDict())

        class badseq(object):
            def __iter__(self):
                return self

            def __next__(self):
                raise Exc()

        self.assertRaises(Exc, {}.update, badseq())

        self.assertRaises(ValueError, {}.update, [(1, 2, 3)])

    def test_fromkeys(self):
        self.assertEqual(dict.fromkeys("abc"), {"a": None, "b": None, "c": None})
        d = Dict()
        self.assertIsNot(d.fromkeys("abc"), d)
        self.assertEqual(d.fromkeys("abc"), {"a": None, "b": None, "c": None})
        self.assertEqual(d.fromkeys((4, 5), 0), {4: 0, 5: 0})
        self.assertEqual(d.fromkeys([]), {})

        def g():
            yield 1

        self.assertEqual(d.fromkeys(g()), {1: None})
        self.assertRaises(TypeError, {}.fromkeys, 3)

        class dictlike(dict):
            pass

        self.assertEqual(dictlike.fromkeys("a"), {"a": None})
        self.assertEqual(dictlike().fromkeys("a"), {"a": None})
        self.assertIsInstance(dictlike.fromkeys("a"), dictlike)
        self.assertIsInstance(dictlike().fromkeys("a"), dictlike)

        class mydict(dict):
            def __new__(cls):
                return Dict()

        ud = mydict.fromkeys("ab")
        self.assertEqual(ud, {"a": None, "b": None})
        self.assertIsInstance(ud, Dict)
        self.assertRaises(TypeError, dict.fromkeys)

        class Exc(Exception):
            pass

        class baddict1(dict):
            def __init__(self):
                raise Exc()

        self.assertRaises(Exc, baddict1.fromkeys, [1])

        class BadSeq(object):
            def __iter__(self):
                return self

            def __next__(self):
                raise Exc()

        self.assertRaises(Exc, dict.fromkeys, BadSeq())

        class baddict2(dict):
            def __setitem__(self, key, value):
                raise Exc()

        self.assertRaises(Exc, baddict2.fromkeys, [1])

        # test fast path for dictionary inputs
        d = dict(zip(range(6), range(6)))
        self.assertEqual(dict.fromkeys(d, 0), dict(zip(range(6), [0] * 6)))

        class baddict3(dict):
            def __new__(cls):
                return d

        d = {i: i for i in range(10)}
        res = d.copy()
        res.update(a=None, b=None, c=None)
        self.assertEqual(baddict3.fromkeys({"a", "b", "c"}), res)

    def test_copy(self):
        d = Dict({1: 1, 2: 2, 3: 3})
        self.assertIsNot(d.copy(), d)
        self.assertEqual(d.copy(), d)
        self.assertEqual(d.copy(), {1: 1, 2: 2, 3: 3})

        copy = d.copy()
        d[4] = 4
        self.assertNotEqual(copy, d)

        self.assertEqual({}.copy(), {})
        self.assertRaises(TypeError, d.copy, None)

    @pytest.mark.slow
    def test_copy_fuzz(self):
        for dict_size in [10, 100, 1000]:  # TODO: 10000, 100000
            dict_size = random.randrange(dict_size // 2, dict_size + dict_size // 2)
            with self.subTest(dict_size=dict_size):
                d = Dict()
                for i in range(dict_size):
                    d[i] = i

                d2 = d.copy()
                self.assertIsNot(d2, d)
                self.assertEqual(d, d2)
                d2["key"] = "value"
                self.assertNotEqual(d, d2)
                self.assertEqual(len(d2), len(d) + 1)

    def test_copy_maintains_tracking(self):
        class A:
            pass

        key = A()

        for d in (Dict(), Dict({"a": 1}), Dict({key: "val"})):
            d2 = d.copy()
            self.assertEqual(gc.is_tracked(d), gc.is_tracked(d2))

    def test_copy_noncompact(self):
        # Dicts don't compact themselves on del/pop operations.
        # Copy will use a slow merging strategy that produces
        # a compacted copy when roughly 33% of dict is a non-used
        # keys-space (to optimize memory footprint).
        # In this test we want to hit the slow/compacting
        # branch of dict.copy() and make sure it works OK.
        d = Dict({k: k for k in range(1000)})
        for k in range(950):
            del d[k]
        d2 = d.copy()
        self.assertEqual(d2, d)

    def test_get(self):
        # We call dict_get because of the conflict with our "get" method
        # from pointers
        d = Dict()
        self.assertIs(d.dict_get("c"), SyNone)
        self.assertEqual(d.dict_get("c", 3), 3)
        d = Dict({"a": 1, "b": 2})
        self.assertIs(d.dict_get("c"), SyNone)
        self.assertEqual(d.dict_get("c", 3), 3)
        self.assertEqual(d.dict_get("a"), 1)
        self.assertEqual(d.dict_get("a", 3), 1)
        self.assertRaises(TypeError, d.get)
        self.assertRaises(TypeError, d.get, None, None, None)

    def test_setdefault(self):
        # dict.setdefault()
        d = Dict()
        self.assertIs(d.setdefault("key0"), SyNone)
        d.setdefault("key0", [])
        self.assertIs(d.setdefault("key0"), SyNone)
        d.setdefault("key", []).append(3)
        self.assertEqual(d["key"][0], 3)
        d.setdefault("key", []).append(4)
        self.assertEqual(len(d["key"]), 2)
        self.assertRaises(TypeError, d.setdefault)

        class Exc(Exception):
            pass

        class BadHash(object):
            fail = False

            def __hash__(self):
                if self.fail:
                    raise Exc()
                else:
                    return 42

        x = BadHash()
        d[x] = 42
        x.fail = True
        self.assertRaises(Exc, d.setdefault, x, [])

    def test_setdefault_atomic(self):
        # Issue #13521: setdefault() calls __hash__ and __eq__ only once.
        class Hashed(object):
            def __init__(self):
                self.hash_count = 0
                self.eq_count = 0

            def __hash__(self):
                self.hash_count += 1
                return 42

            def __eq__(self, other):
                self.eq_count += 1
                return id(self) == id(other)

        hashed1 = Hashed()
        y = {hashed1: 5}
        hashed2 = Hashed()
        y.setdefault(hashed2, [])
        self.assertEqual(hashed1.hash_count, 1)
        self.assertEqual(hashed2.hash_count, 1)
        self.assertEqual(hashed1.eq_count + hashed2.eq_count, 1)

    def test_setitem_atomic_at_resize(self):
        class Hashed(object):
            def __init__(self):
                self.hash_count = 0
                self.eq_count = 0

            def __hash__(self):
                self.hash_count += 1
                return 42

            def __eq__(self, other):
                self.eq_count += 1
                return id(self) == id(other)

        hashed1 = Hashed()
        # 5 items
        y = Dict({hashed1: 5, 0: 0, 1: 1, 2: 2, 3: 3})
        hashed2 = Hashed()
        # 6th item forces a resize
        y[hashed2] = []
        # this is different for UserDict which is 3
        # we are subclassing UserDict so if we match UserDict that should be correct
        # self.assertEqual(hashed1.hash_count, 1)
        self.assertEqual(hashed1.hash_count, 3)
        self.assertEqual(hashed2.hash_count, 1)
        self.assertEqual(hashed1.eq_count + hashed2.eq_count, 1)

    @pytest.mark.slow
    def test_popitem(self):
        # dict.popitem()
        for copymode in -1, +1:
            # -1: b has same structure as a
            # +1: b is a.copy()
            for log2size in range(12):
                size = 2 ** log2size
                a = Dict()
                b = Dict()
                for i in range(size):
                    a[repr(i)] = i
                    if copymode < 0:
                        b[repr(i)] = i
                if copymode > 0:
                    b = a.copy()
                for i in range(size):
                    ka, va = ta = a.popitem()
                    self.assertEqual(va, ka.__int__())
                    kb, vb = tb = b.popitem()
                    self.assertEqual(vb, kb.__int__())
                    self.assertFalse(copymode < 0 and ta != tb)
                self.assertFalse(a)
                self.assertFalse(b)

        d = {}
        self.assertRaises(KeyError, d.popitem)

    def test_pop(self):
        # Tests for pop with specified key
        d = Dict()
        k, v = "abc", "def"
        d[k] = v
        self.assertRaises(KeyError, d.pop, "ghi")

        self.assertEqual(d.pop(k), v)
        self.assertEqual(len(d), 0)

        self.assertRaises(KeyError, d.pop, k)

        self.assertEqual(d.pop(k, v), v)
        d[k] = v
        self.assertEqual(d.pop(k, 1), v)

        self.assertRaises(TypeError, d.pop)

        class Exc(Exception):
            pass

        class BadHash(object):
            fail = False

            def __hash__(self):
                if self.fail:
                    raise Exc()
                else:
                    return 42

        x = BadHash()
        d[x] = 42
        x.fail = True
        self.assertRaises(Exc, d.pop, x)

    def test_mutating_iteration(self):
        # changing dict size during iteration
        d = Dict()
        d[1] = 1
        with self.assertRaises(RuntimeError):
            for i in d:
                d[i + 1] = 1

    def test_mutating_iteration_delete(self):
        # change dict content during iteration
        d = Dict()
        d[0] = 0
        # python 3.8+ raise RuntimeError but older versions do not
        if sys.version_info >= (3, 8):
            with self.assertRaises(RuntimeError):
                for i in d:
                    del d[0]
                    d[0] = 0

    def test_mutating_iteration_delete_over_values(self):
        # change dict content during iteration
        d = Dict()
        d[0] = 0
        # python 3.8+ raise RuntimeError but older versions do not
        if sys.version_info >= (3, 8):
            with self.assertRaises(RuntimeError):
                for i in d.values():
                    del d[0]
                    d[0] = 0

    @pytest.mark.xfail
    def test_mutating_iteration_delete_over_items(self):
        # TODO: proper iterators needed over the views, currently, we convert them to lists
        # change dict content during iteration
        d = Dict()
        d[0] = 0
        if sys.version_info >= (3, 8):
            with self.assertRaises(RuntimeError):
                for i in d.items():
                    del d[0]
                    d[0] = 0

    @pytest.mark.xfail
    def test_mutating_lookup(self):
        # changing dict during a lookup (issue #14417)
        # TODO: investigate this at some point
        class NastyKey:
            mutate_dict = None

            def __init__(self, value):
                self.value = value

            def __hash__(self):
                # hash collision!
                return 1

            def __eq__(self, other):
                if NastyKey.mutate_dict:
                    mydict, key = NastyKey.mutate_dict
                    NastyKey.mutate_dict = None
                    del mydict[key]
                return self.value == other.value

        key1 = NastyKey(1)
        key2 = NastyKey(2)
        d = Dict({key1: 1})
        NastyKey.mutate_dict = (d, key1)
        d[key2] = 2
        self.assertEqual(d, {key2: 2})

    def test_repr(self):
        d = Dict()
        self.assertEqual(repr(d), "{}")
        d[1] = 2
        self.assertEqual(repr(d), "{1: 2}")
        d = Dict()
        d[1] = d
        self.assertEqual(repr(d), "{1: {...}}")

        class Exc(Exception):
            pass

        class BadRepr(object):
            def __repr__(self):
                raise Exc()

        d = Dict({1: BadRepr()})
        self.assertRaises(Exc, repr, d)

    def test_repr_deep(self):
        d = Dict()
        for i in range(sys.getrecursionlimit() + 100):
            d = Dict({1: d})
        self.assertRaises(RecursionError, repr, d)

    def test_eq(self):
        self.assertEqual(Dict(), {})
        self.assertEqual(Dict({1: 2}), {1: 2})

        class Exc(Exception):
            pass

        class BadCmp(object):
            def __eq__(self, other):
                raise Exc()

            def __hash__(self):
                return 1

        d1 = Dict({BadCmp(): 1})
        d2 = Dict({1: 1})

        with self.assertRaises(Exc):
            d1 == d2

    @pytest.mark.xfail
    def test_keys_contained(self):
        self.helper_keys_contained(lambda x: x.keys())
        self.helper_keys_contained(lambda x: x.items())

    @pytest.mark.xfail
    def helper_keys_contained(self, fn):
        # TODO add this when we have set support
        # Test rich comparisons against dict key views, which should behave the
        # same as sets.
        empty = fn(Dict())
        empty2 = fn(Dict())
        smaller = fn(Dict({1: 1, 2: 2}))
        larger = fn(Dict({1: 1, 2: 2, 3: 3}))
        larger2 = fn(Dict({1: 1, 2: 2, 3: 3}))
        larger3 = fn(Dict({4: 1, 2: 2, 3: 3}))

        self.assertTrue(smaller < larger)
        self.assertTrue(smaller <= larger)
        self.assertTrue(larger > smaller)
        self.assertTrue(larger >= smaller)

        self.assertFalse(smaller >= larger)
        self.assertFalse(smaller > larger)
        self.assertFalse(larger <= smaller)
        self.assertFalse(larger < smaller)

        self.assertFalse(smaller < larger3)
        self.assertFalse(smaller <= larger3)
        self.assertFalse(larger3 > smaller)
        self.assertFalse(larger3 >= smaller)

        # Inequality strictness
        self.assertTrue(larger2 >= larger)
        self.assertTrue(larger2 <= larger)
        self.assertFalse(larger2 > larger)
        self.assertFalse(larger2 < larger)

        self.assertTrue(larger == larger2)
        self.assertTrue(smaller != larger)

        # There is an optimization on the zero-element case.
        self.assertTrue(empty == empty2)
        self.assertFalse(empty != empty2)
        self.assertFalse(empty == smaller)
        self.assertTrue(empty != smaller)

        # With the same size, an elementwise compare happens
        self.assertTrue(larger != larger3)
        self.assertFalse(larger == larger3)

    @pytest.mark.xfail
    def test_errors_in_view_containment_check(self):
        # TODO: add support for custom objects
        class C:
            def __eq__(self, other):
                raise RuntimeError

        d1 = Dict({1: C()})
        d2 = Dict({1: C()})
        with self.assertRaises(RuntimeError):
            d1.items() == d2.items()
        with self.assertRaises(RuntimeError):
            d1.items() != d2.items()
        with self.assertRaises(RuntimeError):
            d1.items() <= d2.items()
        with self.assertRaises(RuntimeError):
            d1.items() >= d2.items()

        d3 = Dict({1: C(), 2: C()})
        with self.assertRaises(RuntimeError):
            d2.items() < d3.items()
        with self.assertRaises(RuntimeError):
            d3.items() > d2.items()

    @pytest.mark.xfail
    def test_dictview_set_operations_on_keys(self):
        # TODO add support for sets
        k1 = Dict({1: 1, 2: 2}).keys()
        k2 = Dict({1: 1, 2: 2, 3: 3}).keys()
        k3 = Dict({4: 4}).keys()

        self.assertEqual(k1 - k2, set())
        self.assertEqual(k1 - k3, {1, 2})
        self.assertEqual(k2 - k1, {3})
        self.assertEqual(k3 - k1, {4})
        self.assertEqual(k1 & k2, {1, 2})
        self.assertEqual(k1 & k3, set())
        self.assertEqual(k1 | k2, {1, 2, 3})
        self.assertEqual(k1 ^ k2, {3})
        self.assertEqual(k1 ^ k3, {1, 2, 4})

    @pytest.mark.xfail
    def test_dictview_set_operations_on_items(self):
        # TODO add support for sets
        k1 = Dict({1: 1, 2: 2}).items()
        k2 = Dict({1: 1, 2: 2, 3: 3}).items()
        k3 = Dict({4: 4}).items()

        self.assertEqual(k1 - k2, set())
        self.assertEqual(k1 - k3, {(1, 1), (2, 2)})
        self.assertEqual(k2 - k1, {(3, 3)})
        self.assertEqual(k3 - k1, {(4, 4)})
        self.assertEqual(k1 & k2, {(1, 1), (2, 2)})
        self.assertEqual(k1 & k3, set())
        self.assertEqual(k1 | k2, {(1, 1), (2, 2), (3, 3)})
        self.assertEqual(k1 ^ k2, {(3, 3)})
        self.assertEqual(k1 ^ k3, {(1, 1), (2, 2), (4, 4)})

    @pytest.mark.xfail
    def test_dictview_mixed_set_operations(self):
        # TODO add support for sets
        # Just a few for .keys()
        self.assertTrue(Dict({1: 1}).keys() == {1})
        self.assertEqual(Dict({1: 1}).keys() | {2}, {1, 2})
        # And a few for .items()
        self.assertTrue(Dict({1: 1}).items() == {(1, 1)})

        # This test has been changed to reflect the behavior of UserDict
        self.assertTrue(Dict({(1, 1)}) == {1: 1})

        # UserDict does not support init with set items like:
        # UserDict({2}) so neither do we with Dict
        with pytest.raises(TypeError):
            self.assertEqual(Dict({2}) | Dict({1: 1}).keys(), {1, 2})
            self.assertTrue(Dict({1}) == {1: 1}.keys())
            self.assertEqual(Dict({2}) | Dict({1: 1}).items(), {(1, 1), 2})
            self.assertEqual(Dict({1: 1}).items() | Dict({2}), {(1, 1), 2})

    def test_missing(self):
        # Make sure dict doesn't have a __missing__ method
        self.assertFalse(hasattr(Dict, "__missing__"))
        self.assertFalse(hasattr(Dict(), "__missing__"))
        # Test several cases:
        # (D) subclass defines __missing__ method returning a value
        # (E) subclass defines __missing__ method raising RuntimeError
        # (F) subclass sets __missing__ instance variable (no effect)
        # (G) subclass doesn't define __missing__ at all

        class D(Dict):
            def __missing__(self, key):
                return 42

        d = D({1: 2, 3: 4})
        self.assertEqual(d[1], 2)
        self.assertEqual(d[3], 4)
        self.assertNotIn(2, d)
        self.assertNotIn(2, d.keys())
        self.assertEqual(d[2], 42)

        class E(dict):
            def __missing__(self, key):
                raise RuntimeError(key)

        e = E()
        with self.assertRaises(RuntimeError) as c:
            e[42]
        self.assertEqual(c.exception.args, (42,))

        class F(dict):
            def __init__(self):
                # An instance variable __missing__ should have no effect
                self.__missing__ = lambda key: None

        f = F()
        with self.assertRaises(KeyError) as c:
            f[42]
        self.assertEqual(c.exception.args, (42,))

        class G(dict):
            pass

        g = G()
        with self.assertRaises(KeyError) as c:
            g[42]
        self.assertEqual(c.exception.args, (42,))

    def test_tuple_keyerror(self):
        # SF #1576657
        d = Dict()
        with self.assertRaises(KeyError) as c:
            d[(1,)]
        self.assertEqual(c.exception.args, ((1,),))

    def test_bad_key(self):
        # Dictionary lookups should fail if __eq__() raises an exception.
        class CustomException(Exception):
            pass

        class BadDictKey:
            def __hash__(self):
                return hash(self.__class__)

            def __eq__(self, other):
                if isinstance(other, self.__class__):
                    raise CustomException
                return other

        d = Dict()
        x1 = BadDictKey()
        x2 = BadDictKey()
        d[x1] = 1
        for stmt in [
            "d[x2] = 2",
            "z = d[x2]",
            "x2 in d",
            "d.get(x2)",
            "d.setdefault(x2, 42)",
            "d.pop(x2)",
            "d.update({x2: 2})",
        ]:
            with self.assertRaises(CustomException):
                exec(stmt, locals())

    def test_resize1(self):
        # Dict resizing bug, found by Jack Jansen in 2.2 CVS development.
        # This version got an assert failure in debug build, infinite loop in
        # release build.  Unfortunately, provoking this kind of stuff requires
        # a mix of inserts and deletes hitting exactly the right hash codes in
        # exactly the right order, and I can't think of a randomized approach
        # that would be *likely* to hit a failing case in reasonable time.

        d = Dict()
        for i in range(5):
            d[i] = i
        for i in range(5):
            del d[i]
        for i in range(5, 9):  # i==8 was the problem
            d[i] = i

    def test_resize2(self):
        # Another dict resizing bug (SF bug #1456209).
        # This caused Segmentation faults or Illegal instructions.

        class X(object):
            def __hash__(self):
                return 5

            def __eq__(self, other):
                if resizing:
                    d.clear()
                return False

        d = Dict()
        resizing = False
        d[X()] = 1
        d[X()] = 2
        d[X()] = 3
        d[X()] = 4
        d[X()] = 5
        # now trigger a resize
        resizing = True
        d[9] = 6

    def test_empty_presized_dict_in_freelist(self):
        # Bug #3537: if an empty but presized dict with a size larger
        # than 7 was in the freelist, it triggered an assertion failure
        with self.assertRaises(ZeroDivisionError):
            d = Dict(
                {
                    "a": 1 // 0,
                    "b": None,
                    "c": None,
                    "d": None,
                    "e": None,
                    "f": None,
                    "g": None,
                    "h": None,
                }
            )
            d.clear()

    @pytest.mark.xfail
    @pytest.mark.slow
    def test_container_iterator(self):
        # TODO: make this pass
        # Bug #3680: tp_traverse was not implemented for dictiter and
        # dictview objects.
        class C(object):
            pass

        views = (Dict.items, Dict.values, Dict.keys)
        for v in views:
            obj = C()
            ref = weakref.ref(obj)
            container = {obj: 1}
            obj.v = v(container)
            obj.x = iter(obj.v)
            del obj, container
            gc.collect()
            self.assertIs(ref(), None, "Cycle was not collected")

    def _not_tracked(self, t):
        # Nested containers can take several collections to untrack
        gc.collect()
        gc.collect()
        # UserDict is tracked unlike normal dict so we have to change
        # this test for our Dict
        # self.assertFalse(gc.is_tracked(t), t)
        self.assertTrue(gc.is_tracked(t), t)

    def _tracked(self, t):
        self.assertTrue(gc.is_tracked(t), t)
        gc.collect()
        gc.collect()
        self.assertTrue(gc.is_tracked(t), t)

    @pytest.mark.slow
    @support.cpython_only
    def test_track_literals(self):
        # Test GC-optimization of dict literals
        x, y, z, w = 1.5, "a", (1, None), []

        self._not_tracked(Dict())
        self._not_tracked(Dict({x: (), y: x, z: 1}))
        self._not_tracked(Dict({1: "a", "b": 2}))
        self._not_tracked(Dict({1: 2, (None, True, False, ()): int}))
        self._not_tracked(Dict({1: object()}))

        # Dicts with mutable elements are always tracked, even if those
        # elements are not tracked right now.
        self._tracked(Dict({1: []}))
        self._tracked(Dict({1: ([],)}))
        self._tracked(Dict({1: {}}))
        self._tracked(Dict({1: set()}))

    @pytest.mark.slow
    @support.cpython_only
    def test_track_dynamic(self):
        # Test GC-optimization of dynamically-created dicts
        class MyObject(object):
            pass

        x, y, z, w, o = 1.5, "a", (1, object()), [], MyObject()

        d = Dict()
        self._not_tracked(d)
        d[1] = "a"
        self._not_tracked(d)
        d[y] = 2
        self._not_tracked(d)
        d[z] = 3
        self._not_tracked(d)
        self._not_tracked(d.copy())
        d[4] = w
        self._tracked(d)
        self._tracked(d.copy())
        d[4] = None
        self._not_tracked(d)
        self._not_tracked(d.copy())

        # dd isn't tracked right now, but it may mutate and therefore d
        # which contains it must be tracked.
        d = Dict()
        dd = Dict()
        d[1] = dd
        self._not_tracked(dd)
        self._tracked(d)
        dd[1] = d
        self._tracked(dd)

        d = Dict.fromkeys([x, y, z])
        self._not_tracked(d)
        dd = Dict()
        dd.update(d)
        self._not_tracked(dd)
        d = Dict.fromkeys([x, y, z, o])
        self._tracked(d)
        dd = Dict()
        dd.update(d)
        self._tracked(dd)

        d = Dict(x=x, y=y, z=z)
        self._not_tracked(d)
        d = Dict(x=x, y=y, z=z, w=w)
        self._tracked(d)
        d = Dict()
        d.update(x=x, y=y, z=z)
        self._not_tracked(d)
        d.update(w=w)
        self._tracked(d)

        d = Dict([(x, y), (z, 1)])
        self._not_tracked(d)
        d = Dict([(x, y), (z, w)])
        self._tracked(d)
        d = Dict()
        d.update([(x, y), (z, 1)])
        self._not_tracked(d)
        d.update([(x, y), (z, w)])
        self._tracked(d)

    @support.cpython_only
    def test_track_subtypes(self):
        # Dict subtypes are always tracked
        class MyDict(Dict):
            pass

        self._tracked(MyDict())

    def make_shared_key_dict(self, n):
        class C:
            pass

        dicts = []
        for i in range(n):
            a = C()
            a.x, a.y, a.z = 1, 2, 3
            dicts.append(a.__dict__)

        return dicts

    @support.cpython_only
    def test_splittable_setdefault(self):
        """split table must be combined when setdefault()
        breaks insertion order"""
        a, b = self.make_shared_key_dict(2)

        a["a"] = 1
        size_a = sys.getsizeof(a)
        a["b"] = 2
        b.setdefault("b", 2)
        size_b = sys.getsizeof(b)
        b["a"] = 1

        self.assertGreater(size_b, size_a)
        self.assertEqual(list(a), ["x", "y", "z", "a", "b"])
        self.assertEqual(list(b), ["x", "y", "z", "b", "a"])

    @support.cpython_only
    def test_splittable_del(self):
        """split table must be combined when del d[k]"""
        a, b = self.make_shared_key_dict(2)

        orig_size = sys.getsizeof(a)

        del a["y"]  # split table is combined
        with self.assertRaises(KeyError):
            del a["y"]

        self.assertGreater(sys.getsizeof(a), orig_size)
        self.assertEqual(list(a), ["x", "z"])
        self.assertEqual(list(b), ["x", "y", "z"])

        # Two dicts have different insertion order.
        a["y"] = 42
        self.assertEqual(list(a), ["x", "z", "y"])
        self.assertEqual(list(b), ["x", "y", "z"])

    @support.cpython_only
    def test_splittable_pop(self):
        """split table must be combined when d.pop(k)"""
        a, b = self.make_shared_key_dict(2)

        orig_size = sys.getsizeof(a)

        a.pop("y")  # split table is combined
        with self.assertRaises(KeyError):
            a.pop("y")

        self.assertGreater(sys.getsizeof(a), orig_size)
        self.assertEqual(list(a), ["x", "z"])
        self.assertEqual(list(b), ["x", "y", "z"])

        # Two dicts have different insertion order.
        a["y"] = 42
        self.assertEqual(list(a), ["x", "z", "y"])
        self.assertEqual(list(b), ["x", "y", "z"])

    @support.cpython_only
    def test_splittable_pop_pending(self):
        """pop a pending key in a splitted table should not crash"""
        a, b = self.make_shared_key_dict(2)

        a["a"] = 4
        with self.assertRaises(KeyError):
            b.pop("a")

    @support.cpython_only
    def test_splittable_popitem(self):
        """split table must be combined when d.popitem()"""
        a, b = self.make_shared_key_dict(2)

        orig_size = sys.getsizeof(a)

        item = a.popitem()  # split table is combined
        self.assertEqual(item, ("z", 3))
        with self.assertRaises(KeyError):
            del a["z"]

        self.assertGreater(sys.getsizeof(a), orig_size)
        self.assertEqual(list(a), ["x", "y"])
        self.assertEqual(list(b), ["x", "y", "z"])

    @support.cpython_only
    def test_splittable_setattr_after_pop(self):
        """setattr() must not convert combined table into split table."""
        # Issue 28147
        # third party
        import _testcapi

        class C:
            pass

        a = C()

        a.a = 1
        self.assertTrue(_testcapi.dict_hassplittable(a.__dict__))

        # dict.pop() convert it to combined table
        a.__dict__.pop("a")
        self.assertFalse(_testcapi.dict_hassplittable(a.__dict__))

        # But C should not convert a.__dict__ to split table again.
        a.a = 1
        self.assertFalse(_testcapi.dict_hassplittable(a.__dict__))

        # Same for popitem()
        a = C()
        a.a = 2
        self.assertTrue(_testcapi.dict_hassplittable(a.__dict__))
        a.__dict__.popitem()
        self.assertFalse(_testcapi.dict_hassplittable(a.__dict__))
        a.a = 3
        self.assertFalse(_testcapi.dict_hassplittable(a.__dict__))

    @pytest.mark.xfail
    def test_iterator_pickling(self):
        # set to xfail because we dont really care about pickling
        # see test_valuesiterator_pickling
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            data = Dict({1: "a", 2: "b", 3: "c"})
            it = iter(data)
            d = pickle.dumps(it, proto)
            it = pickle.loads(d)
            self.assertEqual(list(it), list(data))

            it = pickle.loads(d)
            try:
                drop = next(it)
            except StopIteration:
                continue
            d = pickle.dumps(it, proto)
            it = pickle.loads(d)
            del data[drop]
            self.assertEqual(list(it), list(data))

    def test_itemiterator_pickling(self):
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            # UserDict fails these tests so our Dict should fail as well
            with pytest.raises(TypeError):
                data = Dict({1: "a", 2: "b", 3: "c"})
                # dictviews aren't picklable, only their iterators
                itorg = iter(data.items())
                d = pickle.dumps(itorg, proto)
                it = pickle.loads(d)
                # note that the type of the unpickled iterator
                # is not necessarily the same as the original.  It is
                # merely an object supporting the iterator protocol, yielding
                # the same objects as the original one.
                # self.assertEqual(type(itorg), type(it))
                self.assertIsInstance(it, collections.abc.Iterator)
                self.assertEqual(Dict(it), data)

                it = pickle.loads(d)
                drop = next(it)
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                del data[drop[0]]
                self.assertEqual(Dict(it), data)

    def test_valuesiterator_pickling(self):
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            # UserDict fails these tests so our Dict should fail as well
            with pytest.raises(TypeError):
                data = Dict({1: "a", 2: "b", 3: "c"})
                # data.values() isn't picklable, only its iterator
                it = iter(data.values())
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                self.assertEqual(list(it), list(data.values()))

                it = pickle.loads(d)
                drop = next(it)
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                values = list(it) + [drop]
                self.assertEqual(sorted(values), sorted(list(data.values())))

    def test_reverseiterator_pickling(self):
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            # UserDict fails these tests so our Dict should fail as well
            with pytest.raises(TypeError):
                data = Dict({1: "a", 2: "b", 3: "c"})
                it = reversed(data)
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                self.assertEqual(list(it), list(reversed(data)))

                it = pickle.loads(d)
                try:
                    drop = next(it)
                except StopIteration:
                    continue
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                del data[drop]
                self.assertEqual(list(it), list(reversed(data)))

    def test_reverseitemiterator_pickling(self):
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            # UserDict fails these tests so our Dict should fail as well
            with pytest.raises(TypeError):
                data = Dict({1: "a", 2: "b", 3: "c"})
                # dictviews aren't picklable, only their iterators
                itorg = reversed(data.items())
                d = pickle.dumps(itorg, proto)
                it = pickle.loads(d)
                # note that the type of the unpickled iterator
                # is not necessarily the same as the original.  It is
                # merely an object supporting the iterator protocol, yielding
                # the same objects as the original one.
                # self.assertEqual(type(itorg), type(it))
                self.assertIsInstance(it, collections.abc.Iterator)
                self.assertEqual(Dict(it), data)

                it = pickle.loads(d)
                drop = next(it)
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                del data[drop[0]]
                self.assertEqual(Dict(it), data)

    def test_reversevaluesiterator_pickling(self):
        for proto in range(pickle.HIGHEST_PROTOCOL + 1):
            # UserDict fails these tests so our Dict should fail as well
            with pytest.raises(TypeError):
                data = Dict({1: "a", 2: "b", 3: "c"})
                # data.values() isn't picklable, only its iterator
                it = reversed(data.values())
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                self.assertEqual(list(it), list(reversed(data.values())))

                it = pickle.loads(d)
                drop = next(it)
                d = pickle.dumps(it, proto)
                it = pickle.loads(d)
                values = list(it) + [drop]
                self.assertEqual(sorted(values), sorted(data.values()))

    def test_instance_dict_getattr_str_subclass(self):
        class Foo:
            def __init__(self, msg):
                self.msg = msg

        f = Foo("123")

        class _str(str):
            pass

        self.assertEqual(f.msg, getattr(f, _str("msg")))
        self.assertEqual(f.msg, f.__dict__[_str("msg")])

    def test_object_set_item_single_instance_non_str_key(self):
        class Foo:
            pass

        f = Foo()
        f.__dict__[1] = 1
        f.a = "a"
        self.assertEqual(f.__dict__, {1: 1, "a": "a"})

    def check_reentrant_insertion(self, mutate):
        # This object will trigger mutation of the dict when replaced
        # by another value.  Note this relies on refcounting: the test
        # won't achieve its purpose on fully-GCed Python implementations.
        class Mutating:
            def __del__(self):
                mutate(d)

        d = Dict({k: Mutating() for k in "abcdefghijklmnopqr"})
        for k in list(d):
            d[k] = k

    def test_reentrant_insertion(self):
        # Reentrant insertion shouldn't crash (see issue #22653)
        def mutate(d):
            d["b"] = 5

        self.check_reentrant_insertion(mutate)

        def mutate(d):
            d.update(self.__dict__)
            d.clear()

        self.check_reentrant_insertion(mutate)

        def mutate(d):
            while d:
                d.popitem()

        self.check_reentrant_insertion(mutate)

    @pytest.mark.slow
    def test_merge_and_mutate(self):
        # this fails because it expects a RuntimeError when the keys change, however
        # the test_dictitems_contains_use_after_free expects StopIteration when the
        # keys change?
        class X:
            def __hash__(self):
                return 0

            def __eq__(self, o):
                other.clear()
                return False

        l = [(i, 0) for i in range(1, 1337)]
        other = Dict(l)
        other[X()] = 0
        d = Dict({X(): 0, 1: 1})
        self.assertRaises(RuntimeError, d.update, other)

    @pytest.mark.xfail
    @pytest.mark.slow
    def test_free_after_iterating(self):
        # this seems like a bit of a puzzle
        support.check_free_after_iterating(self, iter, Dict)
        support.check_free_after_iterating(self, lambda d: iter(d.keys()), Dict)
        support.check_free_after_iterating(self, lambda d: iter(d.values()), Dict)
        support.check_free_after_iterating(self, lambda d: iter(d.items()), Dict)

    def test_equal_operator_modifying_operand(self):
        # test fix for seg fault reported in bpo-27945 part 3.
        class X:
            def __del__(self):
                dict_b.clear()

            def __eq__(self, other):
                dict_a.clear()
                return True

            def __hash__(self):
                return 13

        dict_a = Dict({X(): 0})
        dict_b = Dict({X(): X()})
        self.assertTrue(dict_a == dict_b)

        # test fix for seg fault reported in bpo-38588 part 1.
        class Y:
            def __eq__(self, other):
                dict_d.clear()
                return True

        dict_c = Dict({0: Y()})
        dict_d = Dict({0: set()})
        self.assertTrue(dict_c == dict_d)

    def test_fromkeys_operator_modifying_dict_operand(self):
        # test fix for seg fault reported in issue 27945 part 4a.
        class X(int):
            def __hash__(self):
                return 13

            def __eq__(self, other):
                if len(d) > 1:
                    d.clear()
                return False

        d = Dict()  # this is required to exist so that d can be constructed!
        d = Dict({X(1): 1, X(2): 2})
        try:
            dict.fromkeys(d)  # shouldn't crash
        except RuntimeError:  # implementation defined
            pass

    def test_fromkeys_operator_modifying_set_operand(self):
        # test fix for seg fault reported in issue 27945 part 4b.
        class X(int):
            def __hash__(self):
                return 13

            def __eq__(self, other):
                if len(d) > 1:
                    d.clear()
                return False

        d = {}  # this is required to exist so that d can be constructed!
        d = {X(1), X(2)}
        try:
            Dict.fromkeys(d)  # shouldn't crash
        except RuntimeError:  # implementation defined
            pass

    @pytest.mark.xfail
    def test_dictitems_contains_use_after_free(self):
        # this seems like a bit of a puzzle
        # see iterator.py for more details
        class X:
            def __eq__(self, other):
                d.clear()
                return NotImplemented

        d = Dict({0: set()})  # TODO: we should be able to support set
        (0, X()) in d.items()

    def test_dict_contain_use_after_free(self):
        # bpo-40489
        class S(str):
            def __eq__(self, other):
                d.clear()
                return NotImplemented

            def __hash__(self):
                return hash("test")

        d = Dict({S(): "value"})
        self.assertFalse("test" in d)

    def test_init_use_after_free(self):
        class X:
            def __hash__(self):
                pair[:] = []
                return 13

        pair = [X(), 123]
        Dict([pair])

    @pytest.mark.xfail
    def test_oob_indexing_dictiter_iternextitem(self):
        class X(int):
            def __del__(self):
                d.clear()

        d = Dict({i: X(i) for i in range(8)})

        def iter_and_mutate():
            for result in d.items():
                if result[0] == 2:
                    d[2] = None  # free d[2] --> X(2).__del__ was called

        self.assertRaises(RuntimeError, iter_and_mutate)

    def test_reversed(self):
        d = Dict({"a": 1, "b": 2, "foo": 0, "c": 3, "d": 4})
        del d["foo"]
        # UserDict does not support reversed so we do not either
        with pytest.raises(TypeError):
            r = reversed(d)
            self.assertEqual(list(r), list("dcba"))
            self.assertRaises(StopIteration, next, r)

    def test_reverse_iterator_for_empty_dict(self):
        # bpo-38525: revered iterator should work properly

        # empty dict is directly used for reference count test
        # UserDict does not support reversed so we do not either
        with pytest.raises(TypeError):
            self.assertEqual(list(reversed(Dict())), [])
            self.assertEqual(list(reversed(Dict().items())), [])
            self.assertEqual(list(reversed(Dict().values())), [])
            self.assertEqual(list(reversed(Dict().keys())), [])

            # dict() and {} don't trigger the same code path
            self.assertEqual(list(reversed(dict())), [])
            self.assertEqual(list(reversed(dict().items())), [])
            self.assertEqual(list(reversed(dict().values())), [])
            self.assertEqual(list(reversed(dict().keys())), [])

    @pytest.mark.xfail
    def test_reverse_iterator_for_shared_shared_dicts(self):
        # UserDict doesnt support reversed and this causes infinite recursion
        # we will just disable this test
        class A:
            def __init__(self, x, y):
                if x:
                    self.x = x
                if y:
                    self.y = y

        self.assertEqual(list(reversed(A(1, 2).__dict__)), ["y", "x"])
        self.assertEqual(list(reversed(A(1, 0).__dict__)), ["x"])
        self.assertEqual(list(reversed(A(0, 1).__dict__)), ["y"])

    @pytest.mark.xfail
    def test_dict_copy_order(self):
        # bpo-34320
        od = collections.OrderedDict([("a", 1), ("b", 2)])
        od.move_to_end("a")
        expected = list(od.items())

        copy = Dict(od)
        self.assertEqual(list(copy.items()), expected)

        # dict subclass doesn't override __iter__
        class CustomDict(Dict):
            pass

        pairs = [("a", 1), ("b", 2), ("c", 3)]

        d = CustomDict(pairs)
        self.assertEqual(pairs, list(Dict(d).items()))

        # UserDict doesnt support reversed and this causes infinite recursion
        # we will just disable this test
        class CustomReversedDict(dict):
            def keys(self):
                return reversed(list(dict.keys(self)))

            __iter__ = keys

            def items(self):
                return reversed(dict.items(self))

        d = CustomReversedDict(pairs)
        self.assertEqual(pairs[::-1], list(dict(d).items()))

    @support.cpython_only
    def test_dict_items_result_gc(self):
        # bpo-42536: dict.items's tuple-reuse speed trick breaks the GC's
        # assumptions about what can be untracked. Make sure we re-track result
        # tuples whenever we reuse them.
        it = iter(Dict({None: []}).items())
        gc.collect()
        # That GC collection probably untracked the recycled internal result
        # tuple, which is initialized to (None, None). Make sure it's re-tracked
        # when it's mutated and returned from __next__:
        self.assertTrue(gc.is_tracked(next(it)))

    @pytest.mark.xfail
    @support.cpython_only
    def test_dict_items_result_gc_reversed(self):
        # UserDict doesnt support reversed and this causes infinite recursion
        # Same as test_dict_items_result_gc above, but reversed.
        it = reversed(Dict({None: []}).items())
        gc.collect()
        self.assertTrue(gc.is_tracked(next(it)))



# third party
import pytest

# syft absolute
import syft as sy


@pytest.mark.vendor(lib="petlib")
def test_ecpt_serde() -> None:
    vm = sy.VirtualMachine()
    client = vm.get_root_client()

    # third party
    import petlib

    sy.load("petlib")

    ec_group = petlib.ec.EcGroup()

    remote_ec_group = ec_group.send(client)
    received_ec_group = remote_ec_group.get()

    assert received_ec_group == ec_group

    ec_pt = petlib.ec.EcPt(ec_group)

    remote_ec_pt = ec_pt.send(client)
    received_ec_group = remote_ec_pt.get()

    assert received_ec_group == ec_pt

    bn = petlib.bn.Bn(1)

    bn_ptr = bn.send(client)
    bn2 = bn_ptr.get()

    assert bn == bn2



# syft absolute
from syft.lib.python.string import String

STRING_PYTHON = "Hello OpenMined"
STRING_SYFT = String(STRING_PYTHON)


def test_id_concat_python_type() -> None:
    val = " 2020"

    result = STRING_SYFT + val

    assert result.id
    assert result.id != STRING_SYFT.id
    assert result == STRING_PYTHON + val

    result = val + STRING_SYFT

    assert result.id
    assert result.id != STRING_SYFT.id
    assert val + STRING_PYTHON == result


def test_id_concat_syft_type() -> None:
    val = String(" 2020")

    result = STRING_SYFT + val

    assert result.id
    assert result.id != STRING_SYFT.id
    assert val.id != result.id
    assert result == STRING_PYTHON + val

    result = val + STRING_SYFT

    assert result.id
    assert result.id != STRING_SYFT.id
    assert val.id != result.id
    assert val + STRING_PYTHON == result



# stdlib
import argparse
from itertools import count

# third party
import gym
import numpy as np
import torch
from torch.distributions import Categorical
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

parser = argparse.ArgumentParser(description="PyTorch REINFORCE example")
parser.add_argument(
    "--gamma",
    type=float,
    default=0.99,
    metavar="G",
    help="discount factor (default: 0.99)",
)
parser.add_argument(
    "--seed",
    type=int,
    default=543,
    metavar="S",
    help="random seed (default: 543)"
)
parser.add_argument(
    "--render",
    action="store_true",
    help="render the environment")
parser.add_argument(
    "--log-interval",
    type=int,
    default=10,
    metavar="N",
    help="interval between training status logs (default: 10)",
)
args = parser.parse_args()


env = gym.make("CartPole-v1")
env.seed(args.seed)
torch.manual_seed(args.seed)


class Policy(nn.Module):
    def __init__(self):
        super(Policy, self).__init__()
        self.affine1 = nn.Linear(4, 128)
        self.dropout = nn.Dropout(p=0.6)
        self.affine2 = nn.Linear(128, 2)

        self.saved_log_probs = []
        self.rewards = []

    def forward(self, x):
        x = self.affine1(x)
        x = self.dropout(x)
        x = F.relu(x)
        action_scores = self.affine2(x)
        return F.softmax(action_scores, dim=1)


policy = Policy()
optimizer = optim.Adam(policy.parameters(), lr=1e-2)
eps = np.finfo(np.float32).eps.item()


def select_action(state):
    state = torch.from_numpy(state).float().unsqueeze(0)
    probs = policy(state)
    m = Categorical(probs)
    action = m.sample()
    policy.saved_log_probs.append(m.log_prob(action))
    return action.item()


def finish_episode():
    R = 0
    policy_loss = []
    returns = []
    for r in policy.rewards[::-1]:
        R = r + args.gamma * R
        returns.insert(0, R)
    returns = torch.tensor(returns)
    returns = (returns - returns.mean()) / (returns.std() + eps)
    for log_prob, R in zip(policy.saved_log_probs, returns):
        policy_loss.append(-log_prob * R)
    optimizer.zero_grad()
    policy_loss = torch.cat(policy_loss).sum()
    policy_loss.backward()
    optimizer.step()
    del policy.rewards[:]
    del policy.saved_log_probs[:]


def main():
    running_reward = 10
    for i_episode in count(1):
        state, ep_reward = env.reset(), 0
        for t in range(1, 10000):  # Don't infinite loop while learning
            action = select_action(state)
            state, reward, done, _ = env.step(action)
            if args.render:
                env.render()
            policy.rewards.append(reward)
            ep_reward += reward
            if done:
                break

        running_reward = 0.05 * ep_reward + (1 - 0.05) * running_reward
        finish_episode()
        if i_episode % args.log_interval == 0:
            print(
                "Episode {}\tLast reward: {:.2f}\tAverage reward: {:.2f}".format(
                    i_episode, ep_reward, running_reward
                )
            )
        if running_reward > env.spec.reward_threshold:
            print(
                "Solved! Running reward is now {} and "
                "the last episode runs to {} time steps!".format(running_reward, t)
            )
            break


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/remote_dataloader/remote_dataset.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/remote_dataloader/remote_dataset.proto",
    package="syft.core.node.common.remote_dataloader",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n1proto/core/remote_dataloader/remote_dataset.proto\x12\'syft.core.node.common.remote_dataloader"0\n\rRemoteDataset\x12\x11\n\tdata_type\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t"v\n\x10RemoteDataLoader\x12N\n\x0eremote_dataset\x18\x01 \x01(\x0b\x32\x36.syft.core.node.common.remote_dataloader.RemoteDataset\x12\x12\n\nbatch_size\x18\x02 \x01(\x03\x62\x06proto3',
)


_REMOTEDATASET = _descriptor.Descriptor(
    name="RemoteDataset",
    full_name="syft.core.node.common.remote_dataloader.RemoteDataset",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="data_type",
            full_name="syft.core.node.common.remote_dataloader.RemoteDataset.data_type",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="path",
            full_name="syft.core.node.common.remote_dataloader.RemoteDataset.path",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=94,
    serialized_end=142,
)


_REMOTEDATALOADER = _descriptor.Descriptor(
    name="RemoteDataLoader",
    full_name="syft.core.node.common.remote_dataloader.RemoteDataLoader",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="remote_dataset",
            full_name="syft.core.node.common.remote_dataloader.RemoteDataLoader.remote_dataset",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="batch_size",
            full_name="syft.core.node.common.remote_dataloader.RemoteDataLoader.batch_size",
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=144,
    serialized_end=262,
)

_REMOTEDATALOADER.fields_by_name["remote_dataset"].message_type = _REMOTEDATASET
DESCRIPTOR.message_types_by_name["RemoteDataset"] = _REMOTEDATASET
DESCRIPTOR.message_types_by_name["RemoteDataLoader"] = _REMOTEDATALOADER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RemoteDataset = _reflection.GeneratedProtocolMessageType(
    "RemoteDataset",
    (_message.Message,),
    {
        "DESCRIPTOR": _REMOTEDATASET,
        "__module__": "proto.core.remote_dataloader.remote_dataset_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.remote_dataloader.RemoteDataset)
    },
)
_sym_db.RegisterMessage(RemoteDataset)

RemoteDataLoader = _reflection.GeneratedProtocolMessageType(
    "RemoteDataLoader",
    (_message.Message,),
    {
        "DESCRIPTOR": _REMOTEDATALOADER,
        "__module__": "proto.core.remote_dataloader.remote_dataset_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.remote_dataloader.RemoteDataLoader)
    },
)
_sym_db.RegisterMessage(RemoteDataLoader)


# @@protoc_insertion_point(module_scope)



# stdlib
import sys
from typing import Optional
from typing import Type

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# syft relative
from ..... import serialize
from .....proto.core.node.common.action.exception_action_pb2 import (
    ExceptionMessage as ExceptionMessage_PB,
)
from .....util import get_fully_qualified_name
from .....util import validate_type
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.deserialize import _deserialize
from ....common.serde.serializable import bind_protobuf
from ....common.uid import UID
from ....io.address import Address


class UnknownPrivateException(Exception):
    pass


@bind_protobuf
@final
class ExceptionMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        msg_id_causing_exception: UID,
        exception_type: Type,
        exception_msg: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.msg_id_causing_exception = msg_id_causing_exception
        self.exception_type = exception_type
        self.exception_msg = exception_msg

    def _object2proto(self) -> ExceptionMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ExceptionMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        # convert exception into fully qualified class path
        fqn = get_fully_qualified_name(obj=self.exception_type)
        module_parts = fqn.split(".")
        _ = module_parts.pop()  # remove incorrect .type ending
        module_parts.append(self.exception_type.__name__)
        fqn = ".".join(module_parts)

        return ExceptionMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            msg_id_causing_exception=serialize(self.msg_id_causing_exception),
            exception_type=fqn,
            exception_msg=self.exception_msg,
        )

    @staticmethod
    def _proto2object(proto: ExceptionMessage_PB) -> "ExceptionMessage":
        """Creates a ExceptionMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of ExceptionMessage
        :rtype: ExceptionMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        # converting fqn of exception class back to real Exception
        module_parts = proto.exception_type.split(".")
        klass = module_parts.pop()
        exception_type = getattr(sys.modules[".".join(module_parts)], klass)

        return ExceptionMessage(
            msg_id=validate_type(_deserialize(blob=proto.msg_id), UID, optional=True),
            address=validate_type(_deserialize(blob=proto.address), Address),
            msg_id_causing_exception=validate_type(
                _deserialize(blob=proto.msg_id_causing_exception), UID
            ),
            exception_type=exception_type,
            exception_msg=proto.exception_msg,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return ExceptionMessage_PB



# stdlib
import asyncio
import atexit
import os
from typing import Any as TypeAny
from typing import Dict as TypeDict
from typing import Optional

# third party
import nest_asyncio

# syft relative
from ...logger import info
from .environment import is_interactive
from .environment import is_jupyter

# Depending on where syft is imported we may need to get the existing event loop or
# create a new one. In the case of Jupyter there is an existing event loop created
# which we need to get otherwise our events might get blocked. If we import via CLI
# there will be no event loop so we need to create one here and set it or the network
# code will try to create its own and in some instances they wont match

try:
    # stdlib
    from asyncio import get_running_loop  # noqa Python >=3.7
except ImportError:  # pragma: no cover
    # stdlib
    from asyncio.events import _get_running_loop as get_running_loop  # pragma: no cover

# first lets get any existing event loop, this will throw an exception in if there is
# no event loop or will return none in older versions of python
loop = None
try:
    loop = get_running_loop()
except RuntimeError:
    pass

# no event loop found, lets create one so that future asyncio imports use the same loop
if loop is None:
    loop = asyncio.new_event_loop()

# set the event loop so its the same everywhere
asyncio._set_running_loop(loop)


# https://github.com/erdewit/nest_asyncio
# patch the event loop to allow nested event loops
nest_asyncio.apply(loop)


def loop_in_thread(loop: TypeAny) -> None:
    asyncio.set_event_loop(loop)
    loop.run_forever()


class EventLoopThread:
    __shared_state: TypeDict[str, TypeAny] = {}

    def __init__(self, loop: TypeAny) -> None:
        if "loop" not in self.__shared_state:
            info("Starting Event Loop")
            self.__shared_state["loop"] = loop
        if "thread" not in self.__shared_state:
            info("Starting Event Loop Thread")
            # daemon=True needed to allow REPL exit() to terminate the thread
            # stdlib
            import threading

            t = threading.Thread(target=loop_in_thread, args=(loop,), daemon=True)
            self.__shared_state["thread"] = t
            t.start()
        self.__dict__ = self.__shared_state

    def shutdown(self) -> None:
        if "loop" in self.__shared_state:
            info("Stopping Event Loop")
            loop.run_until_complete(loop.shutdown_asyncgens())  # type: ignore
            loop.call_soon_threadsafe(loop.stop)  # type: ignore
            del self.__shared_state["loop"]

        if "thread" in self.__shared_state:
            info("Stopping Event Loop Thread")
            t = self.__shared_state["thread"]
            t.join()
            del self.__shared_state["thread"]

    def __del__(self) -> None:
        self.shutdown()


event_loop_thread: Optional[EventLoopThread] = None

# allow threads to be disabled in REPL which will require an event loop to be started
# before running import syft, an example use case is the SCONE container which has
# threads disabled
thread_env = str(os.environ.get("SYFT_USE_EVENT_LOOP_THREAD", "1")).lower()
SYFT_USE_EVENT_LOOP_THREAD = thread_env not in {"0", "false"}

# REPL requires us to create the Thread and Exit handler
if not is_jupyter and is_interactive and SYFT_USE_EVENT_LOOP_THREAD:

    event_loop_thread = EventLoopThread(loop=loop)

    def exit_handler() -> None:
        info("Shutting Down Syft")
        if event_loop_thread is not None:
            event_loop_thread.shutdown()

    atexit.register(exit_handler)

__all__ = ["loop", "event_loop_thread"]



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft relative
from ..... import serialize
from .....logger import debug
from .....logger import error
from .....proto.core.node.common.service.child_node_lifecycle_service_pb2 import (
    RegisterChildNodeMessage as RegisterChildNodeMessage_PB,
)
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.deserialize import _deserialize
from ....common.serde.serializable import bind_protobuf
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from ...common.service.node_service import ImmediateNodeServiceWithoutReply
from .auth import service_auth
from .heritage_update_service import HeritageUpdateMessage


@bind_protobuf
class RegisterChildNodeMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        lookup_id: UID,
        child_node_client_address: Address,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.lookup_id = lookup_id
        self.child_node_client_address = child_node_client_address

    def _object2proto(self) -> RegisterChildNodeMessage_PB:
        debug(f"> {self.icon} -> Proto 🔢")
        return RegisterChildNodeMessage_PB(
            lookup_id=serialize(
                self.lookup_id
            ),  # TODO: not sure if this is needed anymore
            child_node_client_address=serialize(self.child_node_client_address),
            address=serialize(self.address),
            msg_id=serialize(self.id),
        )

    @staticmethod
    def _proto2object(proto: RegisterChildNodeMessage_PB) -> "RegisterChildNodeMessage":
        msg = RegisterChildNodeMessage(
            lookup_id=_deserialize(blob=proto.lookup_id),
            child_node_client_address=_deserialize(
                blob=proto.child_node_client_address
            ),
            address=_deserialize(blob=proto.address),
            msg_id=_deserialize(blob=proto.msg_id),
        )
        debug(f"> {msg.icon} <- 🔢 Proto")
        return msg

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RegisterChildNodeMessage_PB


class ChildNodeLifecycleService(ImmediateNodeServiceWithoutReply):
    @classmethod
    def icon(cls) -> str:
        return "💾"

    @staticmethod
    @service_auth(root_only=True)
    def process(
        node: AbstractNode, msg: RegisterChildNodeMessage, verify_key: VerifyKey
    ) -> None:
        debug(
            f"> Executing {ChildNodeLifecycleService.pprint()} {msg.pprint} on {node.pprint}"
        )
        addr = msg.child_node_client_address
        lookup_id = msg.lookup_id  # TODO: Fix, see above

        node.store[lookup_id] = StorableObject(id=lookup_id, data=addr)

        debug(
            (
                f"> Saving 💾 {addr.pprint} {addr.target_emoji()} with "
                + f"Key: {lookup_id} ➡️ {type(node.store)}"
            )
        )

        # Step 2: update the child node and its descendants with our node.id in their
        # .address objects

        # WARNING: This has no where to go
        # ---------------------------------
        # the goal of this code is to tell the child_node_client, hey you need to update
        # yourself with this new Address (which is up to 4 locations)
        # the old code didn't serialize the incoming object which was actually a real
        # python pointer to the original client
        # now that its a serialized address there are no pointers in memory to the
        # original child clients send_immediate_msg_without_reply function so
        # there is no way to invoke it
        debug(f"> Sending 👪 Update from {node.pprint} back to {addr.target_emoji()}")
        debug("> Update Contains", type(node.address), node.address)
        heritage_msg = HeritageUpdateMessage(
            new_ancestry_address=node.address, address=msg.child_node_client_address
        )

        location = msg.child_node_client_address.target_id.id
        try:
            in_memory_client = node.in_memory_client_registry[location]
            # we need to sign here with the current node not the destination side
            in_memory_client.send_immediate_msg_without_reply(msg=heritage_msg)
            debug(f"> Forwarding {msg.pprint} to {addr.target_emoji()}")
            return None
        except Exception as e:
            error(f"{location} not on nodes in_memory_client. {e}")
            pass
        return None

    @staticmethod
    def message_handler_types() -> List[Type[RegisterChildNodeMessage]]:
        return [RegisterChildNodeMessage]



"""
Define benchmark tests
"""
# stdlib
import atexit
from multiprocessing import Process
from multiprocessing import set_start_method
import os
import time
from typing import Any

# third party
import pytest

# syft absolute
import syft as sy

# syft relative
from ...syft.grid.duet.signaling_server_test import run
from ..pytest_benchmarks.benchmark_send_get_local_test import send_get_list_local
from ..pytest_benchmarks.benchmark_send_get_local_test import send_get_string_local
from ..pytest_benchmarks.benchmark_send_get_multiprocess_test import (
    send_get_list_multiprocess,
)
from ..pytest_benchmarks.benchmark_send_get_multiprocess_test import (
    send_get_string_multiprocess,
)
from ..pytest_benchmarks.benchmark_send_get_multiprocess_test import PORT
from ..pytest_benchmarks.benchmarks_functions_test import list_serde
from ..pytest_benchmarks.benchmarks_functions_test import string_serde

set_start_method("spawn", force=True)

KB = 2 ** 10
MB = 2 ** 20
LIST_TEMPLATE = "a" * (10 * KB)


@pytest.fixture(scope="module")
def signaling_server() -> Process:
    print(f"creating signaling server on port {PORT}")
    grid_proc = Process(target=run, args=(PORT,))
    grid_proc.start()

    def grid_cleanup() -> None:
        print("stop signaling server")
        grid_proc.terminate()
        grid_proc.join()

    atexit.register(grid_cleanup)

    return grid_proc


@pytest.mark.benchmark
@pytest.mark.parametrize("byte_size", [10 * MB, 100 * MB])
def test_string_serde(byte_size: int, benchmark: Any) -> None:
    data = "a" * byte_size
    benchmark.pedantic(string_serde, args=(data,), rounds=3, iterations=3)


@pytest.mark.benchmark
@pytest.mark.parametrize("list_size", [10, 100, 1000])
def test_list_serde(list_size: int, benchmark: Any) -> None:
    data = [LIST_TEMPLATE] * list_size
    benchmark.pedantic(list_serde, args=(data,))


@pytest.mark.benchmark
@pytest.mark.parametrize("byte_size", [10 * KB, 100 * KB, MB, 10 * MB])
def test_duet_string_local(
    byte_size: int, benchmark: Any, root_client: sy.VirtualMachineClient
) -> None:
    data = "a" * byte_size

    benchmark.pedantic(
        send_get_string_local, args=(data, root_client), rounds=3, iterations=3
    )


@pytest.mark.benchmark
@pytest.mark.parametrize("list_size", [10, 100, 1000])
def test_duet_list_local(
    list_size: int, benchmark: Any, root_client: sy.VirtualMachineClient
) -> None:
    data = [LIST_TEMPLATE] * list_size

    benchmark.pedantic(
        send_get_list_local, args=(data, root_client), rounds=3, iterations=3
    )


@pytest.mark.benchmark
@pytest.mark.parametrize("byte_size", [10 * KB, 100 * KB, MB, 10 * MB])
def test_duet_string_multiprocess(
    byte_size: int, benchmark: Any, signaling_server: Process
) -> None:
    time.sleep(3)

    data = "a" * byte_size

    benchmark.pedantic(
        send_get_string_multiprocess, args=(data,), rounds=3, iterations=3
    )


@pytest.mark.benchmark
@pytest.mark.parametrize("list_size", [10, 100, 1000])
def test_duet_list_multiprocess(
    list_size: int, benchmark: Any, signaling_server: Process
) -> None:
    time.sleep(3)

    data = [LIST_TEMPLATE] * list_size

    benchmark.pedantic(send_get_list_multiprocess, args=(data,), rounds=3, iterations=3)


@pytest.mark.skip
@pytest.mark.benchmark
@pytest.mark.parametrize(
    "chunk_size,max_buffer",
    [
        (2 ** 14, 2 ** 18),
        (2 ** 18, 2 ** 23),
        (2 ** 18, 2 ** 24),
        (2 ** 18, 2 ** 25),
    ],
)
def test_duet_chunk_size(
    chunk_size: int, max_buffer: int, benchmark: Any, signaling_server: Process
) -> None:
    time.sleep(3)

    data = "a" * (60 * MB)

    os.environ["DC_MAX_CHUNK_SIZE"] = str(chunk_size)
    os.environ["DC_MAX_BUFSIZE"] = str(max_buffer)

    benchmark.pedantic(
        send_get_string_multiprocess, args=(data,), rounds=2, iterations=2
    )



# stdlib
from unittest.mock import patch

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import pytest
from pytest import mark
from pytest import raises

# syft absolute
from syft import deserialize
from syft import serialize
from syft.core.common import UID
from syft.core.io.address import Address
from syft.core.node.domain import Domain
from syft.core.node.domain.service import RequestMessage


def get_signing_key() -> SigningKey:
    # return a the signing key to use
    key = "e89ff2e651b42393b6ecb5956419088781309d953d72bd73a0968525a3a6a951"
    return SigningKey(bytes.fromhex(key))


def get_verify_key() -> VerifyKey:
    return get_signing_key().verify_key


def test_request_message() -> None:
    addr = Address()
    msg = RequestMessage(
        object_id=UID(),
        address=addr,
        requester_verify_key=get_verify_key(),
        owner_address=addr,
        request_description="test description",
    )

    deserialized_obj = serialize(obj=msg)
    new_obj = deserialize(blob=deserialized_obj)

    assert msg.request_description == new_obj.request_description
    assert msg.address == new_obj.address
    assert msg.owner_address == new_obj.owner_address
    assert msg.object_id == new_obj.object_id
    assert msg.requester_verify_key == get_verify_key()


@pytest.mark.asyncio
@mark.parametrize("method_name", ["accept", "approve"])
def test_accept(method_name: str) -> None:
    node = Domain(name="remote domain")
    node_client = node.get_root_client()

    addr = Address()
    request = RequestMessage(
        object_id=UID(),
        address=addr,
        requester_verify_key=get_verify_key(),
        owner_address=addr,
        owner_client_if_available=node_client,
    )

    with patch.object(
        request.owner_client_if_available, "send_immediate_msg_without_reply"
    ) as mock_send_msg:
        getattr(request, method_name)()
        assert mock_send_msg.call_args[1]["msg"].address == node_client.address
        assert mock_send_msg.call_args[1]["msg"].accept is True
        assert mock_send_msg.call_args[1]["msg"].request_id == request.id


@pytest.mark.asyncio
@mark.parametrize("method_name", ["deny", "reject", "withdraw"])
def test_deny(method_name: str) -> None:
    node = Domain(name="remote domain")
    node_client = node.get_root_client()

    addr = Address()
    request = RequestMessage(
        object_id=UID(),
        address=addr,
        requester_verify_key=get_verify_key(),
        owner_address=addr,
        owner_client_if_available=node_client,
    )

    with patch.object(
        request.owner_client_if_available, "send_immediate_msg_without_reply"
    ) as mock_send_msg:
        getattr(request, method_name)()
        assert mock_send_msg.call_args[1]["msg"].address == node_client.address
        assert mock_send_msg.call_args[1]["msg"].accept is False
        assert mock_send_msg.call_args[1]["msg"].request_id == request.id


def test_fail_accept_request_message() -> None:
    addr = Address()
    request = RequestMessage(
        object_id=UID(),
        address=addr,
        requester_verify_key=get_verify_key(),
        owner_address=addr,
    )

    with raises(Exception, match="No way to dispatch Accept Message."):
        request.accept()


def test_fail_deny_request_message() -> None:
    addr = Address()
    request = RequestMessage(
        object_id=UID(),
        address=addr,
        requester_verify_key=get_verify_key(),
        owner_address=addr,
    )

    with raises(Exception, match="No way to dispatch Deny Message."):
        request.deny()


def test_fail_process_request_service() -> None:
    addr = Address()
    request = RequestMessage(
        object_id=UID(),
        address=addr,
        requester_verify_key=get_verify_key(),
        owner_address=addr,
    )

    with raises(Exception, match="No way to dispatch Deny Message."):
        request.deny()



# stdlib
from typing import Any
from typing import Type

# third party
from google.protobuf.message import Message
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from ....logger import traceback_and_raise
from ....util import random_name


def bind_protobuf(cls: Any) -> Any:
    protobuf_schema = cls.get_protobuf_schema()
    # If protobuf already has schema2type, means it's related to multiple types.
    # Set it's schema2type to None, becuase we can't take use of it anymore.
    if hasattr(protobuf_schema, "schema2type"):
        protobuf_schema.schema2type = None
    else:
        protobuf_schema.schema2type = cls

    return cls


class Serializable:
    """When we want a custom object to be serializable within the Syft ecosystem
    (as outline in the tutorial above), the first thing we need to do is have it
    subclass from this class. You must then do the following in order for the
    subclass to be properly implemented:

    - implement a protobuf file in the "PySyft/proto" folder for this custom class.
    - compile the protobuf file by running `bash scripts/build_proto`
    - find the generated python file in syft.proto
    - import the generated protobuf class into my custom class
    - implement get_protobuf_schema
    - implement <my class>._object2proto() method to serialize the object to protobuf
    - implement <my class>._proto2object() to deserialize the protobuf object

    At this point, your class should be ready to serialize and deserialize! Don't
    forget to add tests for your object!

    If you want to wrap an existing type (like a torch.tensor) to be used in our serialization
    ecosystem, you should consider wrapping it. Wrapping means that we NEVER use the wrapper
    further more into our ecosystem, we only need an easy interface to serialize wrappers.

    Eg:

    .. code-block:: python

        class WrapperInt(Serializable)
            def __init__(self, value: int):
               self.int_obj = value

            def _object2proto(self) -> WrapperIntPB:
               ...

            @staticmethod
            def _proto2object(proto) -> int:
               ...

            @staticmethod
            def get_protobuf_schema() -> GeneratedProtocolMessageType:
               ...

            @staticmethod
            def get_wrapped_type() -> type:
               return int

    You must implement the following in order for the subclass to be properly implemented to be
    seen as a wrapper:

    - everything presented in the first tutorial of this docstring.
    - implement get_wrapped_type to return the wrapped type.

    Note: A wrapper should NEVER be used in the codebase, these are only for serialization purposes
    on external objects.

    After doing all of the above steps, you can call something like sy.serialize(5) and be
    serialized using our messaging proto backbone.
    """

    @property
    def named(self) -> str:
        if hasattr(self, "name"):
            return self.name  # type: ignore
        else:
            return "UNNAMED"

    @property
    def class_name(self) -> str:
        return str(self.__class__.__name__)

    @property
    def icon(self) -> str:
        # as in cereal, get it!?
        return "🌾"

    @property
    def pprint(self) -> str:
        return f"{self.icon} {self.named} ({self.class_name})"

    @staticmethod
    def _proto2object(proto: Message) -> "Serializable":
        """This method converts a protobuf object into a subclass of Serializable

        This method must be implemented for all classes which subclassSerializable - namely
        all classes which can be serialized within the Syft ecosystem. It should convert the
        corresponding protobuf message for the subclass into an instance of the class. This
        allows all the logic which goes from protobuf message to other formats (JSON, binary, etc.)
        to be generic and simply inherited from the rest of this class.

        :param proto: the protobuf object to be converted into an instance of type(self)
        :param type: Message
        :return: an instance of type(self)
        :rtype: Serializable

        """
        traceback_and_raise(NotImplementedError)

    def _object2proto(self) -> Message:
        """This methods converts self into a protobuf object

        This method must be implemented by all subclasses so that generic high-level functions
        implemented here (such as serialize(, to_bytes=True), etc) know how to convert the object into
        a protobuf object before further converting it into the requested format.

        :return: a protobuf message
        :rtype: Message
        """

        traceback_and_raise(NotImplementedError)
        raise NotImplementedError

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        traceback_and_raise(NotImplementedError)

    @staticmethod
    def get_wrapped_type() -> Type:
        """
        This static method returns the wrapped type, if the current class is
        a wrapper over an external object.

        :return: the wrapped type
        :rtype: type
        """
        traceback_and_raise(NotImplementedError)

    @staticmethod
    def random_name() -> str:
        return random_name()



# stdlib
import sys
from typing import Any

# third party
import pytest

# syft absolute
from syft.lib.python.int import Int

VALID_UNDERSCORE_LITERALS = [
    "0_0_0",
    "4_2",
    "1_0000_0000",
    "0b1001_0100",
    "0xffff_ffff",
    "0o5_7_7",
    "1_00_00.5",
    "1_00_00.5e5",
    "1_00_00e5_1",
    "1e1_0",
    ".1_4",
    ".1_4e1",
    "0b_0",
    "0x_f",
    "0o_5",
    "1_00_00j",
    "1_00_00.5j",
    "1_00_00e5_1j",
    ".1_4j",
    "(1_2.5+3_3j)",
    "(.5_6j)",
]

INVALID_UNDERSCORE_LITERALS = [
    # Trailing underscores:
    "0_",
    "42_",
    "1.4j_",
    "0x_",
    "0b1_",
    "0xf_",
    "0o5_",
    "0 if 1_Else 1",
    # Underscores in the base selector:
    "0_b0",
    "0_xf",
    "0_o5",
    # Old-style octal, still disallowed:
    "0_7",
    "09_99",
    # Multiple consecutive underscores:
    "4_______2",
    "0.1__4",
    "0.1__4j",
    "0b1001__0100",
    "0xffff__ffff",
    "0x___",
    "0o5__77",
    "1e1__0",
    "1e1__0j",
    # Underscore right before a dot:
    "1_.4",
    "1_.4j",
    # Underscore right after a dot:
    "1._4",
    "1._4j",
    "._5",
    "._5j",
    # Underscore right after a sign:
    "1.0e+_1",
    "1.0e+_1j",
    # Underscore right before j:
    "1.4_j",
    "1.4e5_j",
    # Underscore right before e:
    "1_e1",
    "1.4_e1",
    "1.4_e1j",
    # Underscore right after e:
    "1e_1",
    "1.4e_1",
    "1.4e_1j",
    # Complex cases with parens:
    "(1+1.5_j_)",
    "(1+1.5_j)",
]

L = [
    ("0", 0),
    ("1", 1),
    ("9", 9),
    ("10", 10),
    ("99", 99),
    ("100", 100),
    ("314", 314),
    (" 314", 314),
    ("314 ", 314),
    ("  \t\t  314  \t\t  ", 314),
    (repr(sys.maxsize), sys.maxsize),
    ("  1x", ValueError),
    ("  1  ", 1),
    ("  1\02  ", ValueError),
    ("", ValueError),
    (" ", ValueError),
    ("  \t\t  ", ValueError),
    ("\u0200", ValueError),
]


class IntSubclass(Int):
    pass


def test_basic() -> None:
    assert Int(314) == 314
    assert Int(3.14) == 3
    # Check that conversion from float truncates towards zero
    assert Int(-3.14) == -3
    assert Int(3.9) == 3
    assert Int(-3.9) == -3
    assert Int(3.5) == 3
    assert Int(-3.5) == -3
    assert Int("-3") == -3
    assert Int(" -3 ") == -3
    assert Int("\N{EM SPACE}-3\N{EN SPACE}") == -3
    # Different base:
    assert Int("10", 16) == 16
    # Test conversion from strings and various anomalies
    for s, v in L:
        for sign in "", "+", "-":
            for prefix in "", " ", "\t", "  \t\t  ":
                ss = prefix + sign + s
                vv = v
                if sign == "-" and v is not ValueError:
                    vv = -v
                try:
                    assert Int(ss) == vv
                except ValueError:
                    pass

    s = repr(-1 - sys.maxsize)
    x = Int(s)
    assert Int(x + 1) == -sys.maxsize
    assert isinstance(x, Int)
    # should return Int
    assert Int(s[1:]) == sys.maxsize + 1

    # should return Int
    x = Int(1e100)
    assert isinstance(x, Int)
    x = Int(-1e100)
    assert isinstance(x, Int)

    # SF bug 434186:  0x80000000/2 != 0x80000000>>1.
    # Worked by accident in Windows release build, but failed in debug build.
    # Failed in all Linux builds.
    x = -1 - sys.maxsize
    assert Int(x >> 1) == x // 2

    x = Int("1" * 600)
    assert isinstance(x, Int)

    # TODO: check why is this not working
    # with pytest.raises(TypeError):
    #     Int(1, 12)

    assert Int("0o123", 0) == 83
    assert Int("0x123", 16) == 291

    # Bug 1679: "0x" is not a valid hex literal
    with pytest.raises(ValueError):
        Int("0x", 16)
    with pytest.raises(ValueError):
        Int("0x", 0)

    with pytest.raises(ValueError):
        Int("0o", 8)
    with pytest.raises(ValueError):
        Int("0o", 0)

    with pytest.raises(ValueError):
        Int("0b", 2)
    with pytest.raises(ValueError):
        Int("0b", 0)

    # SF bug 1334662: Int(string, base) wrong answers
    # Various representations of 2**32 evaluated to 0
    # rather than 2**32 in previous versions

    assert Int("100000000000000000000000000000000", 2) == 4294967296
    assert Int("102002022201221111211", 3) == 4294967296
    assert Int("10000000000000000", 4) == 4294967296
    assert Int("32244002423141", 5) == 4294967296
    assert Int("1550104015504", 6) == 4294967296
    assert Int("211301422354", 7) == 4294967296
    assert Int("40000000000", 8) == 4294967296
    assert Int("12068657454", 9) == 4294967296
    assert Int("4294967296", 10) == 4294967296
    assert Int("1904440554", 11) == 4294967296
    assert Int("9ba461594", 12) == 4294967296
    assert Int("535a79889", 13) == 4294967296
    assert Int("2ca5b7464", 14) == 4294967296
    assert Int("1a20dcd81", 15) == 4294967296
    assert Int("100000000", 16) == 4294967296
    assert Int("a7ffda91", 17) == 4294967296
    assert Int("704he7g4", 18) == 4294967296
    assert Int("4f5aff66", 19) == 4294967296
    assert Int("3723ai4g", 20) == 4294967296
    assert Int("281d55i4", 21) == 4294967296
    assert Int("1fj8b184", 22) == 4294967296
    assert Int("1606k7ic", 23) == 4294967296
    assert Int("mb994ag", 24) == 4294967296
    assert Int("hek2mgl", 25) == 4294967296
    assert Int("dnchbnm", 26) == 4294967296
    assert Int("b28jpdm", 27) == 4294967296
    assert Int("8pfgih4", 28) == 4294967296
    assert Int("76beigg", 29) == 4294967296
    assert Int("5qmcpqg", 30) == 4294967296
    assert Int("4q0jto4", 31) == 4294967296
    assert Int("4000000", 32) == 4294967296
    assert Int("3aokq94", 33) == 4294967296
    assert Int("2qhxjli", 34) == 4294967296
    assert Int("2br45qb", 35) == 4294967296
    assert Int("1z141z4", 36) == 4294967296

    # tests with base 0
    # this fails on 3.0, but in 2.x the old octal syntax is allowed
    assert Int(" 0o123  ", 0) == 83
    assert Int(" 0o123  ", 0) == 83
    assert Int("000", 0) == 0
    assert Int("0o123", 0) == 83
    assert Int("0x123", 0) == 291
    assert Int("0b100", 0) == 4
    assert Int(" 0O123   ", 0) == 83
    assert Int(" 0X123  ", 0) == 291
    assert Int(" 0B100 ", 0) == 4

    # without base still base 10
    assert Int("0123") == 123
    assert Int("0123", 10) == 123

    # tests with prefix and base != 0
    assert Int("0x123", 16) == 291
    assert Int("0o123", 8) == 83
    assert Int("0b100", 2) == 4
    assert Int("0X123", 16) == 291
    assert Int("0O123", 8) == 83
    assert Int("0B100", 2) == 4

    # the code has special checks for the first character after the
    #  type prefix
    with pytest.raises(ValueError):
        Int("0b2", 2)
    with pytest.raises(ValueError):
        Int("0b02", 2)
    with pytest.raises(ValueError):
        Int("0B2", 2)
    with pytest.raises(ValueError):
        Int("0B02", 2)
    with pytest.raises(ValueError):
        Int("0o8", 8)
    with pytest.raises(ValueError):
        Int("0o08", 8)
    with pytest.raises(ValueError):
        Int("0O8", 8)
    with pytest.raises(ValueError):
        Int("0O08", 8)
    with pytest.raises(ValueError):
        Int("0xg", 16)
    with pytest.raises(ValueError):
        Int("0x0g", 16)
    with pytest.raises(ValueError):
        Int("0Xg", 16)
    with pytest.raises(ValueError):
        Int("0X0g", 16)

    # SF bug 1334662: Int(string, base) wrong answers
    # Checks for proper evaluation of 2**32 + 1
    assert Int("100000000000000000000000000000001", 2) == 4294967297
    assert Int("102002022201221111212", 3) == 4294967297
    assert Int("10000000000000001", 4) == 4294967297
    assert Int("32244002423142", 5) == 4294967297
    assert Int("1550104015505", 6) == 4294967297
    assert Int("211301422355", 7) == 4294967297
    assert Int("40000000001", 8) == 4294967297
    assert Int("12068657455", 9) == 4294967297
    assert Int("4294967297", 10) == 4294967297
    assert Int("1904440555", 11) == 4294967297
    assert Int("9ba461595", 12) == 4294967297
    assert Int("535a7988a", 13) == 4294967297
    assert Int("2ca5b7465", 14) == 4294967297
    assert Int("1a20dcd82", 15) == 4294967297
    assert Int("100000001", 16) == 4294967297
    assert Int("a7ffda92", 17) == 4294967297
    assert Int("704he7g5", 18) == 4294967297
    assert Int("4f5aff67", 19) == 4294967297
    assert Int("3723ai4h", 20) == 4294967297
    assert Int("281d55i5", 21) == 4294967297
    assert Int("1fj8b185", 22) == 4294967297
    assert Int("1606k7id", 23) == 4294967297
    assert Int("mb994ah", 24) == 4294967297
    assert Int("hek2mgm", 25) == 4294967297
    assert Int("dnchbnn", 26) == 4294967297
    assert Int("b28jpdn", 27) == 4294967297
    assert Int("8pfgih5", 28) == 4294967297
    assert Int("76beigh", 29) == 4294967297
    assert Int("5qmcpqh", 30) == 4294967297
    assert Int("4q0jto5", 31) == 4294967297
    assert Int("4000001", 32) == 4294967297
    assert Int("3aokq95", 33) == 4294967297
    assert Int("2qhxjlj", 34) == 4294967297
    assert Int("2br45qc", 35) == 4294967297
    assert Int("1z141z5", 36) == 4294967297


@pytest.mark.xfail
def test_underscores() -> None:
    for lit in VALID_UNDERSCORE_LITERALS:
        if any(ch in lit for ch in ".eEjJ"):
            continue
        assert Int(lit, 0) == eval(lit)
        # TODO: check why is this not working
        assert Int(lit, 0), Int(lit.replace("_", "") == 0)
    for lit in INVALID_UNDERSCORE_LITERALS:
        if any(ch in lit for ch in ".eEjJ"):
            continue
        with pytest.raises(ValueError):
            Int(lit, 0)
    # Additional test cases with bases != 0, only for the constructor:
    assert Int("1_00", 3) == 9
    assert Int("0_100") == 100  # not valid as a literal!
    assert Int(b"1_00") == 100  # byte underscore
    with pytest.raises(ValueError):
        Int("_100")
    with pytest.raises(ValueError):
        Int("+_100")
    with pytest.raises(ValueError):
        Int("1__00")
    with pytest.raises(ValueError):
        Int("100_")


def test_small_Ints() -> None:
    # Bug #3236: Return small longs from PyLong_FromString
    # TODO: These are not going to work if we use is, brainstorm for a workaround
    assert Int("10") == 10
    assert Int("-1") == -1
    assert Int(b"10") == 10
    assert Int(b"-1") == -1


def test_no_args() -> None:
    assert Int() == 0


@pytest.mark.xfail
def test_keyword_args() -> None:
    # Test invoking Int() using keyword arguments.
    assert Int("100", base=2) == 4
    with pytest.raises(TypeError):
        Int(x=1.2)
    with pytest.raises(TypeError):
        Int(x="100", base=2)
    # TODO these should fail
    with pytest.raises(TypeError):
        Int(base=10)
    with pytest.raises(TypeError):
        Int(base=0)


def test_Int_base_limits() -> None:
    """Testing the supported limits of the Int() base parameter."""
    assert Int("0", 5) == 0
    with pytest.raises(ValueError):
        Int("0", 1)
    with pytest.raises(ValueError):
        Int("0", 37)
    with pytest.raises(ValueError):
        Int("0", -909)  # An old magic value base from Python 2.
    with pytest.raises(ValueError):
        Int("0", base=0 - (2 ** 234))
    with pytest.raises(ValueError):
        Int("0", base=2 ** 234)
    # Bases 2 through 36 are supported.
    for base in range(2, 37):
        assert Int("0", base=base) == 0


def test_Int_base_bad_types() -> None:
    """Not Integer types are not valid bases; issue16772."""
    with pytest.raises(TypeError):
        Int("0", 5.5)
    with pytest.raises(TypeError):
        Int("0", 5.0)


def test_Int_base_indexable() -> None:
    class MyIndexable(object):
        def __init__(self, value: Any):
            self.value = value

        def __index__(self):
            return self.value

    # Check out of range bases.
    for base in 2 ** 100, -(2 ** 100), 1, 37:
        with pytest.raises(ValueError):
            Int("43", base)

    # Check in-range bases.
    assert Int("101", base=MyIndexable(2)) == 5
    assert Int("101", base=MyIndexable(10)) == 101
    assert Int("101", base=MyIndexable(36)) == 1 + 36 ** 2


def test_Int_memoryview() -> None:
    assert Int(memoryview(b"123")[1:3]) == 23
    assert Int(memoryview(b"123\x00")[1:3]) == 23
    assert Int(memoryview(b"123 ")[1:3]) == 23
    assert Int(memoryview(b"123A")[1:3]) == 23
    assert Int(memoryview(b"1234")[1:3]) == 23


def test_string_float() -> None:
    with pytest.raises(ValueError):
        Int("1.2")


@pytest.mark.xfail
def test_Intconversion() -> None:
    # Test __Int__()
    class ClassicMissingMethods:
        pass

    with pytest.raises(TypeError):
        Int(ClassicMissingMethods())

    class MissingMethods(object):
        pass

    with pytest.raises(TypeError):
        Int(MissingMethods())

    class Foo0:
        def __Int__(self):
            return 42

    # TODO this should work
    assert Int(Foo0()) == 42

    class Classic:
        pass

    for base in (object, Classic):

        class IntOverridesTrunc(base):
            def __Int__(self):
                return 42

            def __trunc__(self):
                return -12

        # TODO this should work
        assert Int(IntOverridesTrunc()) == 42

        class JustTrunc(base):
            def __trunc__(self):
                return 42

        # TODO this should work
        assert Int(JustTrunc()) == 42

        class ExceptionalTrunc(base):
            def __trunc__(self):
                1 / 0

        with pytest.raises(ZeroDivisionError):
            Int(ExceptionalTrunc())

        for trunc_result_base in (object, Classic):

            class Index(trunc_result_base):
                def __index__(self):
                    return 42

            class TruncReturnsNonInt(base):
                def __trunc__(self):
                    return Index()

            # TODO this should work
            assert Int(TruncReturnsNonInt()) == 42

            class Intable(trunc_result_base):
                def __Int__(self):
                    return 42

            class TruncReturnsNonIndex(base):
                def __trunc__(self):
                    return Intable()

            # TODO this should work
            assert Int(TruncReturnsNonInt()) == 42

            class NonIntegral(trunc_result_base):
                def __trunc__(self):
                    # Check that we avoid infinite recursion.
                    return NonIntegral()

            class TruncReturnsNonIntegral(base):
                def __trunc__(self):
                    return NonIntegral()

            try:
                Int(TruncReturnsNonIntegral())
            except TypeError as e:
                assert str(e) == "__trunc__ returned non-Integral" " (type NonIntegral)"
            else:
                raise Exception("Failed to raise TypeError with %s")

            # Regression test for bugs.python.org/issue16060.
            class BadInt(trunc_result_base):
                def __Int__(self):
                    return 42.0

            class TruncReturnsBadInt(base):
                def __trunc__(self):
                    return BadInt()

            with pytest.raises(TypeError):
                Int(TruncReturnsBadInt())


def test_Int_subclass_with_index() -> None:
    class MyIndex(Int):
        def __index__(self):
            return 42

    class BadIndex(Int):
        def __index__(self):
            return 42.0

    my_Int = MyIndex(7)
    assert Int(my_Int) == 7
    assert Int(my_Int) == 7

    assert Int(BadIndex()) == 0


@pytest.mark.xfail
def test_Int_subclass_with_Int() -> None:
    class MyInt(Int):
        def __Int__(self):
            return 42

    class BadInt(Int):
        def __Int__(self):
            return 42.0

    # TODO this should work
    my_Int = MyInt(7)
    assert Int(my_Int) == 7
    assert Int(my_Int) == 42

    # TODO this should work
    my_Int = BadInt(7)
    assert Int(my_Int) == 7
    with pytest.raises(TypeError):
        Int(my_Int)


@pytest.mark.xfail
def test_Int_returns_Int_subclass() -> None:
    class BadIndex:
        def __index__(self):
            return True

    class BadIndex2(Int):
        def __index__(self):
            return True

    class BadInt:
        def __Int__(self):
            return True

    class BadInt2(Int):
        def __Int__(self):
            return True

    class TruncReturnsBadIndex:
        def __trunc__(self):
            return BadIndex()

    class TruncReturnsBadInt:
        def __trunc__(self):
            return BadInt()

    class TruncReturnsIntSubclass:
        def __trunc__(self):
            return True

    # TODO: this should work
    bad_Int = BadIndex()
    with pytest.raises(DeprecationWarning):
        n = Int(bad_Int)
    assert Int(n) == 1
    # this is not going to work
    assert type(n) is Int

    bad_Int = BadIndex2()
    n = Int(bad_Int)
    assert Int(n) == 0

    # TODO: this should work
    bad_Int = BadInt()
    with pytest.raises(DeprecationWarning):
        n = Int(bad_Int)
    assert Int(n) == 1
    # not going to work
    # self.assertIs(type(n), Int)

    # TODO: this should work
    bad_Int = BadInt2()
    with pytest.raises(DeprecationWarning):
        n = Int(bad_Int)
    assert Int(n) == 1
    # not going to work
    # self.assertIs(type(n), Int)


def test_error_message() -> None:
    def check(s, base=None):
        with pytest.raises(ValueError):
            if base is None:
                Int(s)
            else:
                Int(s, base)

    check("\xbd")
    check("123\xbd")
    check("  123 456  ")

    check("123\x00")
    # SF bug 1545497: embedded NULs were not detected with explicit base
    check("123\x00", 10)
    check("123\x00 245", 20)
    check("123\x00 245", 16)
    check("123\x00245", 20)
    check("123\x00245", 16)
    # byte string with embedded NUL
    check(b"123\x00")
    check(b"123\x00", 10)
    # non-UTF-8 byte string
    check(b"123\xbd")
    check(b"123\xbd", 10)
    # lone surrogate in Unicode string
    check("123\ud800")
    check("123\ud800", 10)


def test_protobof_schema():
    assert Int.get_protobuf_schema()


def test_bytes():
    assert (
        Int.from_bytes(Int(42).to_bytes(4, "big", signed=True), "big", signed=True)
        == 42
    )
    assert (
        Int.from_bytes(Int(-42).to_bytes(4, "big", signed=True), "big", signed=True)
        == -42
    )



# syft absolute
from syft.lib.python.int import Int

test_int = Int(10)
other = Int(2)

python_int = 10


def test_id_abs():
    res = test_int.__abs__()
    py_res = python_int.__abs__()

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_add():
    res = test_int.__add__(other)
    py_res = python_int.__add__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_and():
    res = test_int.__and__(other)
    py_res = python_int.__and__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_ceil():
    res = test_int.__ceil__()
    py_res = python_int.__ceil__()

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_divmod():
    q, r = test_int.__divmod__(other)

    assert q.id
    assert r.id


def test_id_eq():
    res = test_int.__eq__(other)
    py_res = python_int.__eq__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_float():
    res = test_int.__float__()
    py_res = python_int.__float__()

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_floor():
    res = test_int.__floor__()
    py_res = python_int.__floor__()

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_floordiv():
    res = test_int.__floordiv__(other)
    py_res = python_int.__floordiv__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_ge():
    res = test_int.__ge__(other)
    py_res = python_int.__ge__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_gt():
    res = test_int.__gt__(other)
    py_res = python_int.__gt__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_hash():
    res = test_int.__hash__()
    py_res = python_int.__hash__()

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_iadd():
    res = test_int.__iadd__(other)

    assert res.id
    assert res.id == test_int.id


def test_id_ifloordiv():
    res = test_int.__ifloordiv__(other)

    assert res.id
    assert res.id == test_int.id


def test_id_imod():
    res = test_int.__imod__(other)

    assert res.id
    assert res.id == test_int.id


def test_id_imul():
    res = test_int.__imul__(other)

    assert res.id
    assert res.id == test_int.id


def test_id_invert():
    res = test_int.__invert__()
    py_res = python_int.__invert__()

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_ipow():
    res = test_int.__ipow__(other)

    assert res.id
    assert res.id == test_int.id


def test_id_isub():
    res = test_int.__isub__(other)

    assert res.id
    assert res.id == test_int.id


def test_id_itruediv():
    res = test_int.__itruediv__(other)

    assert res.id
    assert res.id == test_int.id


def test_id_le():
    res = test_int.__le__(other)
    py_res = python_int.__le__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_lshift():
    res = test_int.__lshift__(other)
    py_res = python_int.__lshift__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_lt():
    res = test_int.__lt__(other)
    py_res = python_int.__lt__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_mod():
    res = test_int.__mod__(other)
    py_res = python_int.__mod__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_mul():
    res = test_int.__mul__(other)
    py_res = python_int.__mul__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id


def test_id_ne():
    res = test_int.__ne__(other)
    py_res = python_int.__ne__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_int.id



# stdlib
import os
import zipfile

# PyTorch 1.1 moves _download_url_to_file
#   from torch.utils.model_zoo to torch.hub
# PyTorch 1.0 exists another _download_url_to_file
#   2 argument
# TODO: If you remove support PyTorch 1.0 or older,
#       You should remove torch.utils.model_zoo
#       Ref. PyTorch #18758
#         https://github.com/pytorch/pytorch/pull/18758/commits
try:
    # third party
    from torch.utils.model_zoo import _download_url_to_file
except ImportError:
    try:
        # third party
        from torch.hub import download_url_to_file as _download_url_to_file
    except ImportError:
        # third party
        from torch.hub import _download_url_to_file


def unzip(source_filename, dest_dir):
    with zipfile.ZipFile(source_filename) as zf:
        zf.extractall(path=dest_dir)


if __name__ == "__main__":
    _download_url_to_file(
        "https://www.dropbox.com/s/lrvwfehqdcxoza8/saved_models.zip?dl=1",
        "saved_models.zip",
        None,
        True,
    )
    unzip("saved_models.zip", ".")



"""
 Syft's AST submodule is responsible for remote call executions. Basically, an AST is a
 tree that maps function calls to their exact path and knows what to do with that node in the tree.

 Example: Suppose we want to append an object to a List. This means that we need to
 know where we could find the `append` method, so we need to know the following chain:

                globals  <- the global scope or the entry point of execution (hidden)
                   |
                  syft   <- the syft module
                   |
                  lib   <- a submodule of syft
                   |
                  List <- the class we were looking for
                   |
                append <- the method we were looking for

When performing remote execution, this lookup path has to be resolved as well. This is where
the AST submodule comes in handy. AST is responsible for:

A. Remote execution.
B. Local execution.

A. Remote Execution
Remote execution can be performed only when an AST has been constructed with a client. Check
syft/core/node to be familiar with the roles of clients and nodes. Each valid action on the AST
triggers an Action (GetSetStaticAttributeAction, GetSetPropertyAction, etc). This kind of actions
requires: The path on resolving the required node (on the above example, the path is
`syft.lib.List.append`), the object on which to perform it (given by the `__self` attribute and from
`id_at_location` if executed on a pointer) and the attributes if needed.

B. Local Execution
After a call has been made, the remote execution starts. For this, we need a local handler for the result
of the remote execution through a `Pointer`. The AST is responsible for generating all the permitted methods
and attributes on a `Pointer` and the return type of the performed action.


The existing types of nodes are:
* a `Globals`, which is the entry point of an execution, from which point on we can only access Modules.

* a `Callable`, which can be a node for a method, a static method, a function, or a constructor. This
node can no longer have any attributes.

* a `Class`, which is a node that represents a Python Class. This node can contain methods - Callable,
static methods - Callable, class methods - Callable, slot attributes - StaticAttribute, properties - Property,
enum attributes - EnumAttribute.

* a `Module`, which represents a Python file/module. This node can have attributes such as global
variables - StaticAttribute, global functions - Callable and classes - Klass.

* a `StaticAttribute` - represents attribute of a Class or of a Module. This node cannot have any attributes.
This can be remotely get and set.

* a `Property` - represents a @property object of a class. This node cannot have any attributes, which means
that this node is essentially a "leaf". This node can perform get and set remotely.

* an `EnumAttribute` - represents the fields generated by an `Enum`. This node cannot have any attributes,
meaning that this node is essentially a "leaf". This node can perform get remotely.
"""

# stdlib
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple
from typing import Union

# syft relative
from . import attribute  # noqa: F401
from . import callable  # noqa: F401
from . import enum  # noqa: F401
from . import globals  # noqa: F401
from . import klass  # noqa: F401
from . import module  # noqa: F401
from . import property  # noqa: F401
from . import static_attr  # noqa: F401


def get_parent(
    path: str, root: Union[attribute.Attribute, globals.Globals, module.Module]
) -> Union[module.Module, klass.Class]:
    """
    Return the parent of a given path.

    Args:
        path: The full path to an object.
        root: The collection of frameworks held in the global namespace.

    Returns:
        The parent module or class.

    Examples:
        For instance, given the syft project root directory, the parent to the path
        `syft.lib.python.Int` is `python`.
    """
    parent = root
    for step in path.split(".")[:-1]:
        if step in parent.attrs:
            parent = parent.attrs[step]

    if not isinstance(parent, (module.Module, klass.Class)):
        raise ValueError(f"Expected (Module, Class), but got {type(parent)}")

    return parent


def add_modules(
    ast: globals.Globals,
    modules: Union[TypeList[str], TypeList[TypeTuple[str, TypeAny]]],
) -> None:
    """
    Parse a list of modules and register each module to its corresponding parent object in the AST path.

    Args:
        ast: The global AST.
        modules: A list of modules, either a path in string format or a tuple of the path in string and a reference.
    """
    for mod in modules:
        # In case reference is present
        if isinstance(mod, tuple):
            target_module, ref = mod
        else:
            target_module, ref = mod, None

        parent = get_parent(target_module, ast)
        attr_name = target_module.rsplit(".", 1)[-1]
        parent.add_attr(
            attr_name=attr_name,
            attr=module.Module(
                path_and_name=target_module,
                object_ref=ref,
                return_type_name="",
                client=ast.client,
            ),
        )


def add_classes(
    ast: globals.Globals,
    paths: TypeList[TypeTuple[str, str, TypeAny]],
) -> None:
    """
    Parse a list of classes and register each class to its corresponding parent object in the AST path.

    Args:
        ast: The global AST.
        paths: A list of classes, each of which is a tuple of the path, the return type, and its reference.
    """
    for path, return_type, ref in paths:
        parent = get_parent(path, ast)
        attr_name = path.rsplit(".", 1)[-1]
        parent.add_attr(
            attr_name=attr_name,
            attr=klass.Class(
                path_and_name=path,
                object_ref=ref,
                return_type_name=return_type,
                client=ast.client,
                parent=parent,
            ),
        )


def add_methods(
    ast: globals.Globals,
    paths: TypeList[TypeTuple[str, str]],
) -> None:
    """
    Parse a list of methods and register each method to its corresponding parent object in the AST path.

    Args:
        ast: The global AST.
        paths: A list of methods, each of which is a tuple of the method's path and its return type.
    """
    for path, return_type in paths:
        parent = get_parent(path, ast)
        path_list = path.split(".")
        parent.add_path(
            path=path_list,
            index=len(path_list) - 1,
            return_type_name=return_type,
        )



# stdlib
from typing import Any
from typing import Optional

# syft relative
from ...core.common.uid import UID
from ...logger import traceback_and_raise
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet


class Iterator(PyPrimitive):
    def __init__(self, _ref: Any, max_len: Optional[int] = None):
        super().__init__()
        self._obj_ref = _ref
        self._index = 0
        self._id = UID()
        self.max_len = max_len
        self.exhausted = False

    def __iter__(self) -> "Iterator":
        return self

    def __len__(self) -> int:
        try:
            return len(self._obj_ref)
        except Exception as e:
            traceback_and_raise(e)

    def __reduce__(self) -> Any:
        # see these tests: test_valuesiterator_pickling and test_iterator_pickling
        raise TypeError(f"Pickling {type(self)} is not supported.")

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        if hasattr(other, "_obj_ref"):
            res = self._obj_ref == other._obj_ref
        else:
            res = self._obj_ref == other

        return PrimitiveFactory.generate_primitive(value=res)

    def __next__(self) -> Any:
        # we need to do lots of getattr / setattr because some times the __next__
        # method gets called with a generator
        try:
            if hasattr(self, "_obj_ref"):
                _obj_ref = self._obj_ref
            else:
                # we got handed a generator directly into __next__
                # happens in test_reversed_iterator
                _obj_ref = self

            # max_len == None means the _ref could update while iterating. While that
            # shouldn't happen with a IteratorPointer, it can happen on a local Iterator.
            # If thats the case we just calculate it each time. Importantly we need to
            # still set exhausted otherwise the test case in list_test.py wont pass.
            max_len = None
            if hasattr(self, "max_len"):
                max_len = self.max_len

            if max_len is None:
                try:
                    if hasattr(_obj_ref, "__len__"):
                        max_len = _obj_ref.__len__()
                except AttributeError:
                    # I am not sure why this happens on some types
                    pass

            exhausted = getattr(self, "exhausted", False)
            self_index = getattr(self, "_index", 0)
            if (max_len is not None and self_index >= max_len) or exhausted:
                setattr(self, "exhausted", True)
                raise StopIteration

            try:
                if hasattr(_obj_ref, "__next__"):
                    try:
                        obj = next(_obj_ref)
                    except Exception as e:
                        if type(e) is StopIteration:
                            raise e
                        if type(e) is AttributeError:
                            # no _mapping exhausted?
                            raise StopIteration()
                        if type(e) is NameError:
                            # free after use?
                            raise StopIteration()

                        # test_dictitems_contains_use_after_free wants us to StopIteration
                        # test_merge_and_mutate and test_mutating_iteration wants us to
                        # raise a RuntimeError
                        # see:
                        # def test_dictitems_contains_use_after_free(self):
                        # Lets RuntimeError for now
                        raise RuntimeError

                elif hasattr(_obj_ref, "__getitem__") and hasattr(self, "_index"):
                    obj = _obj_ref[self._index]
                elif hasattr(_obj_ref, "__iter__"):
                    # collections.abc.* KeysView, ValuesView, ItemsView end up here
                    # they do not have __next__ or __getitem__ but they do have __iter__
                    # so we can just replace our self._obj_ref and keep going
                    setattr(self, "_obj_ref", _obj_ref.__iter__())
                    # obj = next(self._obj_ref) # just call self.__next__() instead
                    return self.__next__()
                else:
                    raise ValueError("Can't iterate through given object.")
            except StopIteration as e:
                setattr(self, "exhausted", True)
                raise e

            if hasattr(self, "_index"):
                self._index += 1
            return obj
        except Exception as e:
            raise e

    def upcast(self) -> Any:
        return iter(self._obj_ref)
